import asyncio
import os
import signal
import threading
import time as ttime
import types
from collections import defaultdict
from functools import partial
from traceback import FrameSummary, extract_tb

import pytest
from event_model import DocumentNames

from bluesky import Msg, RunEngine
from bluesky.plan_stubs import (
    abs_set,
    checkpoint,
    configure,
    declare_stream,
    pause,
    trigger_and_read,
    wait,
    wait_for,
)
from bluesky.plans import count, grid_scan, scan
from bluesky.preprocessors import (
    SupplementalData,
    baseline_wrapper,
    finalize_wrapper,
    reset_positions_decorator,
    rewindable_wrapper,
    run_decorator,
    run_wrapper,
    subs_wrapper,
)
from bluesky.protocols import Status
from bluesky.run_engine import (
    FailedStatus,
    IllegalMessageSequence,
    NoReplayAllowed,
    RequestAbort,
    RequestStop,
    RunEngineInterrupted,
    RunEngineStateMachine,
    TransitionError,
    WaitForTimeoutError,
)
from bluesky.tests import requires_ophyd, uses_os_kill_sigint
from bluesky.tests.utils import DocCollector, MsgCollector

from .utils import _careful_event_set, _fabricate_asycio_event


def test_states():
    assert RunEngineStateMachine.States.states() == [
        "idle",
        "running",
        "pausing",
        "paused",
        "halting",
        "stopping",
        "aborting",
        "suspending",
        "panicked",
    ]


def test_panic_trap(RE):
    RE._state = "panicked"
    for k in RunEngineStateMachine.States.states():
        if k != "panicked":
            with pytest.raises(TransitionError):
                RE._state = k


def test_state_is_readonly(RE):
    with pytest.raises(AttributeError):
        RE.state = "running"


@pytest.mark.parametrize(
    "deferred_pause_delay, is_pause_set",
    [
        (None, False),  # Do not pause
        (0.1, False),  # Pause before the checkpoint
        (1.4, True),  # Pause after the checkpoint
    ],
)
def test_deferred_pause_requested(RE, deferred_pause_delay, is_pause_set):
    """
    Test for ``deferred_pause_requested``.
    """
    assert RE.deferred_pause_requested is False

    plan = [Msg("sleep", None, 1), Msg("checkpoint"), Msg("sleep", None, 1)]

    pause_req_immediate = None

    def _pause():
        # This function is requesting the deferred pause and reads the property
        nonlocal pause_req_immediate
        RE.request_pause(defer=True)
        ttime.sleep(0.1)
        pause_req_immediate = RE.deferred_pause_requested

    t = None
    if deferred_pause_delay is not None:
        t = threading.Timer(deferred_pause_delay, _pause)
        t.start()

    try:
        RE(plan)
    except Exception:
        pass

    if deferred_pause_delay is not None:
        assert pause_req_immediate is True

    assert RE.deferred_pause_requested is is_pause_set
    if t is not None:
        t.join()


def test_verbose(RE, hw):
    RE.verbose = True
    assert RE.verbose
    # Emit all four kinds of document, exercising the logging.
    RE(
        [
            Msg("open_run"),
            Msg("declare_stream", None, hw.det, name="primary"),
            Msg("create", name="primary"),
            Msg("read", hw.det),
            Msg("save"),
            Msg("close_run"),
        ]
    )


def test_reset(RE):
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("open_run"), Msg("pause")])

    assert len(RE._run_bundlers) > 0
    RE.reset()
    assert len(RE._run_bundlers) == 0


def test_running_from_paused_state_raises(RE):
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("pause")])
    assert RE.state == "paused"
    with pytest.raises(RuntimeError):
        RE([Msg("null")])
    RE.resume()
    assert RE.state == "idle"
    RE([Msg("null")])


def test_resuming_from_idle_state_raises(RE):
    with pytest.raises(RuntimeError):
        RE.resume()
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("pause")])
    assert RE.state == "paused"
    RE.resume()
    assert RE.state == "idle"
    with pytest.raises(RuntimeError):
        RE.resume()


def test_stopping_from_idle_state_raises(RE):
    with pytest.raises(TransitionError):
        RE.stop()


def test_pausing_from_idle_state_raises(RE):
    with pytest.raises(TransitionError):
        RE.request_pause(defer=False)


def test_aborting_from_idle_state_raises(RE):
    with pytest.raises(TransitionError):
        RE.abort()


def test_register(RE):
    mutable = {}
    RE.verbose = True

    async def func(msg):
        mutable["flag"] = True

    def plan():
        yield Msg("custom-command")

    RE.register_command("custom-command", func)
    RE(plan())
    assert "flag" in mutable
    # Unregister command; now the Msg should not be recognized.
    RE.unregister_command("custom-command")
    with pytest.raises(KeyError):
        RE([Msg("custom-command")])


def test_stop_motors_and_log_any_errors(RE, hw):
    # test that if stopping one motor raises an error, we can carry on
    stopped = {}

    def stop(self, *, success=False):
        stopped[self.name] = True

    def stop_encounters_error(self, *, success=False):
        stopped[self.name] = True
        raise Exception

    motor = hw.motor1
    broken_motor = hw.motor2
    motor.stop = types.MethodType(stop, motor)
    broken_motor.stop = types.MethodType(stop_encounters_error, broken_motor)

    with pytest.raises(RunEngineInterrupted):
        RE([Msg("set", broken_motor, 1), Msg("set", motor, 1), Msg("pause")])
    assert "motor1" in stopped
    assert "motor2" in stopped
    RE.stop()

    with pytest.raises(RunEngineInterrupted):
        RE([Msg("set", motor, 1), Msg("set", broken_motor, 1), Msg("pause")])
    assert "motor1" in stopped
    assert "motor2" in stopped
    RE.stop()


@requires_ophyd
def test_unstage_and_log_errors(RE):
    unstaged = {}

    from ophyd.sim import SynAxis

    class MoverWithFlag(SynAxis):
        def stage(self):
            return [self]

        def unstage(self):
            unstaged[self.name] = True
            return [self]

    class BrokenMoverWithFlag(SynAxis):
        def stage(self):
            return [self]

        def unstage(self):
            unstaged[self.name] = True
            return [self]

    a = MoverWithFlag(name="a")
    b = BrokenMoverWithFlag(name="b")

    unstaged.clear()
    RE([Msg("stage", a), Msg("stage", b)])
    assert "a" in unstaged
    assert "b" in unstaged

    unstaged.clear()
    RE([Msg("stage", b), Msg("stage", a)])
    assert "a" in unstaged
    assert "b" in unstaged


def test_open_run_twice_is_illegal(RE):
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("open_run"), Msg("open_run")])


def test_saving_without_an_open_bundle_is_illegal(RE):
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("open_run"), Msg("save")])


def test_dropping_without_an_open_bundle_is_illegal(RE):
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("open_run"), Msg("drop")])


def test_opening_a_bundle_without_a_run_is_illegal(RE):
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("create", name="primary")])


def test_checkpoint_inside_a_bundle_is_illegal(RE):
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("open_run"), Msg("create", name="primary"), Msg("checkpoint")])


def test_redundant_monitors_are_illegal(RE):
    class Dummy:
        def __init__(self, name):
            self.name = name

        def read_configuration(self):
            return {}

        def describe_configuration(self):
            return {}

        def describe(self):
            return {}

        def read(self):
            return {}

        def subscribe(self, *args, **kwargs):
            pass

        def clear_sub(self, *args, **kwargs):
            pass

    dummy = Dummy("dummy")

    with pytest.raises(IllegalMessageSequence):
        RE([Msg("open_run"), Msg("monitor", dummy), Msg("monitor", dummy)])

    # Monitoring, unmonitoring, and monitoring again is legal.
    RE([Msg("open_run"), Msg("monitor", dummy), Msg("unmonitor", dummy), Msg("monitor", dummy)])

    # Monitoring outside a run is illegal.
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("monitor", dummy)])

    # Unmonitoring something that was never monitored is illegal.
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("unmonitor", dummy)])


def test_empty_bundle(RE, hw):
    mutable = {}

    def cb(name, doc):
        mutable["flag"] = True

    # In this case, an Event should be emitted.
    mutable.clear()
    RE(
        [
            Msg("open_run"),
            Msg("declare_stream", None, hw.det, name="primary"),
            Msg("create", name="primary"),
            Msg("read", hw.det),
            Msg("save"),
        ],
        {"event": cb},
    )
    assert "flag" in mutable

    # In this case, an Event should not be emitted because the bundle is
    # emtpy (i.e., there are no readings.)
    mutable.clear()
    RE(
        [Msg("open_run"), Msg("declare_stream", None, name="primary"), Msg("create", name="primary"), Msg("save")],
        {"event": cb},
    )
    assert "flag" not in mutable


def test_dispatcher_unsubscribe_all(RE):
    def count_callbacks(RE):
        return sum([len(cbs) for cbs in RE.dispatcher.cb_registry.callbacks.values()])

    def cb(name, doc):
        pass

    RE.subscribe(cb)
    assert count_callbacks(RE) == len(DocumentNames)
    RE.dispatcher.unsubscribe_all()
    assert count_callbacks(RE) == 0


def test_stage_and_unstage_are_optional_methods(RE):
    class Dummy:
        pass

    dummy = Dummy()

    RE([Msg("stage", dummy), Msg("unstage", dummy)])


def test_need_both_stage_and_unstage_for_stagable(RE):
    class StageOnly:
        def stage(self):
            raise Exception

    class UnstageOnly:
        def unstage(self):
            raise Exception

    stage_only = StageOnly()
    unstage_only = UnstageOnly()
    RE(
        [
            Msg("stage", stage_only),
            Msg("stage", unstage_only),
            Msg("unstage", stage_only),
            Msg("unstage", unstage_only),
        ]
    )


def test_pause_resume_devices(RE):
    paused = {}
    resumed = {}

    class Dummy:
        def __init__(self, name):
            self.name = name

        def pause(self):
            paused[self.name] = True

        def resume(self):
            resumed[self.name] = True

    dummy = Dummy("dummy")

    with pytest.raises(RunEngineInterrupted):
        RE([Msg("stage", dummy), Msg("pause")])
    RE.resume()
    assert "dummy" in paused
    assert "dummy" in resumed


def test_stage_and_unstage_status_objects(RE):
    from ophyd import StatusBase

    staged = {}
    unstaged = {}
    started_stage = {}
    started_unstage = {}

    class Dummy:
        def __init__(self, name: str) -> None:
            self.name: str = name

        def _callback(self, d: dict, st: Status):
            d[self.name] = True
            st.set_finished()  # type: ignore

        def stage(self) -> Status:
            st = StatusBase()
            started_stage[self.name] = True
            threading.Timer(1, self._callback, args=[staged, st]).start()
            return st

        def unstage(self) -> Status:
            st = StatusBase()
            started_unstage[self.name] = True
            threading.Timer(1, self._callback, args=[unstaged, st]).start()
            return st

    dummy1 = Dummy("d1")
    dummy2 = Dummy("d2")

    def my_plan():
        yield Msg("stage", dummy1, group="stage_group")
        yield Msg("stage", dummy2, group="stage_group")

        assert "d1" in started_stage and "d2" in started_stage
        assert not staged

        yield Msg("wait", group="stage_group")
        assert "d1" in staged and "d2" in staged

        yield Msg("unstage", dummy1, group="unstage_group")
        yield Msg("unstage", dummy2, group="unstage_group")

        assert "d1" in started_unstage and "d2" in started_unstage
        assert not unstaged

        yield Msg("wait", group="unstage_group")
        assert "d1" in unstaged and "d2" in unstaged

    start = ttime.monotonic()
    RE(my_plan())
    stop = ttime.monotonic()

    assert 2 <= stop - start < 3


def test_bad_call_args(RE):
    with pytest.raises(Exception):  # noqa: B017
        RE(53)
    assert RE.state == "idle"


def test_record_interruptions(RE):
    docs = defaultdict(lambda: [])

    def collect(name, doc):
        print("HI", name)
        docs[name].append(doc)
        print(docs)

    RE.subscribe(collect)
    RE.ignore_callback_exceptions = False
    RE.msg_hook = print

    # The 'pause' inside the run should generate an event iff
    # record_interruptions is True.
    plan = [Msg("pause"), Msg("open_run"), Msg("pause"), Msg("close_run")]

    assert not RE.record_interruptions
    with pytest.raises(RunEngineInterrupted):
        RE(plan)
    with pytest.raises(RunEngineInterrupted):
        RE.resume()
    RE.resume()
    assert len(docs["descriptor"]) == 0
    assert len(docs["event"]) == 0

    RE.record_interruptions = True
    with pytest.raises(RunEngineInterrupted):
        RE(plan)
    with pytest.raises(RunEngineInterrupted):
        RE.resume()
    RE.resume()
    assert len(docs["descriptor"]) == 1
    assert len(docs["event"]) == 2
    docs["event"][0]["data"]["interruption"] == "pause"  # noqa: B015
    docs["event"][1]["data"]["interruption"] == "resume"  # noqa: B015


@requires_ophyd
def _make_unrewindable_marker():
    from ophyd.sim import SynAxis, SynGauss

    class UnReplayableSynGauss(SynGauss):
        def pause(self):
            raise NoReplayAllowed()

    motor = SynAxis(name="motor")

    def test_plan(motor, det):
        yield Msg("set", motor, 0)
        yield Msg("trigger", det)
        yield Msg("pause")
        yield Msg("set", motor, 1)
        yield Msg("trigger", det)

    inps = []
    inps.append(
        (
            test_plan,
            motor,
            UnReplayableSynGauss("det", motor, "motor", center=0, Imax=1),
            ["set", "trigger", "pause", "set", "trigger"],
        )
    )

    inps.append(
        (
            test_plan,
            motor,
            SynGauss("det", motor, "motor", center=0, Imax=1),
            ["set", "trigger", "pause", "set", "trigger", "set", "trigger"],
        )
    )

    return pytest.mark.parametrize("plan,motor,det,msg_seq", inps)


@_make_unrewindable_marker()
def test_unrewindable_det(RE, plan, motor, det, msg_seq):
    msgs = []

    def collector(msg):
        msgs.append(msg)

    RE.msg_hook = collector
    with pytest.raises(RunEngineInterrupted):
        RE(plan(motor, det))
    RE.resume()
    assert [m.command for m in msgs] == msg_seq


@requires_ophyd
def _make_unrewindable_suspender_marker():
    from ophyd.sim import SynAxis, SynGauss

    class UnReplayableSynGauss(SynGauss):
        def pause(self):
            raise NoReplayAllowed()

    motor = SynAxis(name="motor")

    def test_plan(motor, det):
        yield Msg("set", motor, 0)
        yield Msg("trigger", det)
        yield Msg("sleep", None, 1)
        yield Msg("set", motor, 0)
        yield Msg("trigger", det)

    inps = []
    inps.append(
        (
            test_plan,
            motor,
            UnReplayableSynGauss("det", motor, "motor", center=0, Imax=1),
            [
                "set",
                "trigger",
                "sleep",
                "_start_suspender",
                "rewindable",
                "wait_for",
                "_resume_from_suspender",
                "rewindable",
                "set",
                "trigger",
            ],
        )
    )

    inps.append(
        (
            test_plan,
            motor,
            SynGauss("det", motor, "motor", center=0, Imax=1),
            [
                "set",
                "trigger",
                "sleep",
                "_start_suspender",
                "rewindable",
                "wait_for",
                "_resume_from_suspender",
                "rewindable",
                "set",
                "trigger",
                "sleep",
                "set",
                "trigger",
            ],
        )
    )

    return pytest.mark.parametrize("plan,motor,det,msg_seq", inps)


@_make_unrewindable_suspender_marker()
def test_unrewindable_det_suspend(RE, plan, motor, det, msg_seq):
    from bluesky.utils import ts_msg_hook

    msgs = []
    loop = RE.loop

    def collector(msg):
        ts_msg_hook(msg)
        msgs.append(msg)

    RE.msg_hook = collector

    ev = _fabricate_asycio_event(loop)

    timer = threading.Timer(0.5, RE.request_suspend, kwargs=dict(fut=ev.wait))  # noqa: C408
    timer.start()

    def verbose_set():
        print("seting")
        ev.set()

    loop.call_soon_threadsafe(loop.call_later, 1, verbose_set)

    RE(plan(motor, det))
    assert [m.command for m in msgs] == msg_seq
    timer.join()


@pytest.mark.parametrize("unpause_func", [lambda RE: RE.stop(), lambda RE: RE.abort(), lambda RE: RE.resume()])
def test_cleanup_after_pause(RE, unpause_func, hw):
    motor = hw.motor
    motor.set(1024)

    @reset_positions_decorator()
    def simple_plan(motor):
        for j in range(15):
            yield Msg("set", motor, j)
        yield Msg("pause")
        for j in range(15):
            yield Msg("set", motor, -j)

    with pytest.raises(RunEngineInterrupted):
        RE(simple_plan(motor))
    assert motor.position == 14
    unpause_func(RE)
    assert motor.position == 1024


@pytest.mark.parametrize(
    "unpause_func,excp",
    [
        (lambda RE: RE.stop(), RequestStop),
        (lambda RE: RE.abort(), RequestAbort),
        (lambda RE: RE.halt(), GeneratorExit),
    ],
)
def test_exit_raise(RE, unpause_func, excp):
    flag = False

    @reset_positions_decorator()
    def simple_plan():
        nonlocal flag
        try:
            yield from checkpoint()
            yield Msg("pause")
        except excp:
            flag = True

    with pytest.raises(RunEngineInterrupted):
        RE(simple_plan())
    unpause_func(RE)
    assert flag


@uses_os_kill_sigint
def test_sigint_three_hits(RE, hw):
    import time

    motor = hw.motor
    motor.delay = 0.5

    pid = os.getpid()

    def sim_kill(n):
        for j in range(n):  # noqa: B007
            time.sleep(0.05)
            os.kill(pid, signal.SIGINT)

    lp = RE.loop
    motor.loop = lp

    def self_sig_int_plan():
        threading.Timer(0.05, sim_kill, (3,)).start()
        yield from abs_set(motor, 1, wait=True)

    start_time = ttime.time()
    with pytest.raises(RunEngineInterrupted):
        RE(finalize_wrapper(self_sig_int_plan(), abs_set(motor, 0, wait=True)))
    end_time = ttime.time()
    # not enough time for motor to cleanup, but long enough to start
    assert 0.05 < end_time - start_time < 0.2
    RE.abort()  # now cleanup

    done_cleanup_time = ttime.time()
    # this should be 0.5 (the motor.delay) above, leave sloppy for CI
    assert 0.3 < done_cleanup_time - end_time < 0.6


@uses_os_kill_sigint
def test_sigint_many_hits_pln(RE):
    pid = os.getpid()

    def sim_kill(n):
        for j in range(n):
            print("KILL", j)
            ttime.sleep(0.05)
            os.kill(pid, signal.SIGINT)

    def hanging_plan():
        "a plan that blocks the RunEngine's normal Ctrl+C handing with a sleep"
        for j in range(100):  # noqa: B007
            ttime.sleep(0.1)
        yield Msg("null")

    start_time = ttime.time()
    timer = threading.Timer(0.2, sim_kill, (11,))
    timer.start()
    with pytest.raises(RunEngineInterrupted):
        RE(hanging_plan())
    # Check that hammering SIGINT escaped from that 10-second sleep.
    assert ttime.time() - start_time < 5
    # The KeyboardInterrupt will have been converted to a hard pause that
    # the test plan can not handle so we abort and go to idle.
    assert RE.state == "idle"
    timer.join()


@uses_os_kill_sigint
def test_sigint_many_hits_panic(RE):
    raise pytest.skip("hangs tests on exit")
    pid = os.getpid()

    def sim_kill(n):
        for j in range(n):
            print("KILL", j, ttime.monotonic() - start_time)
            ttime.sleep(0.05)
            os.kill(pid, signal.SIGINT)

    def hanging_plan():
        "a plan that blocks the RunEngine's normal Ctrl+C handing with a sleep"
        yield Msg("null")
        ttime.sleep(5)
        yield Msg("null")

    start_time = ttime.monotonic()
    timer = threading.Timer(0.2, sim_kill, (11,))
    timer.start()
    with pytest.raises(RunEngineInterrupted):
        RE(hanging_plan())
    # Check that hammering SIGINT escaped from that 5-second sleep.
    assert (ttime.monotonic() - start_time) < 2.5
    # The KeyboardInterrupt but because we could not shut down, panic!
    assert RE.state == "panicked"
    timer.join()

    with pytest.raises(RuntimeError):
        RE([])

    with pytest.raises(RuntimeError):
        RE.stop()

    with pytest.raises(RuntimeError):
        RE.halt()

    with pytest.raises(RuntimeError):
        RE.abort()

    with pytest.raises(RuntimeError):
        RE.resume()

    with pytest.raises(RuntimeError):
        RE.request_pause()


@uses_os_kill_sigint
def test_sigint_many_hits_cb(RE):
    pid = os.getpid()

    def sim_kill(n):
        for j in range(n):  # noqa: B007
            print("KILL")
            ttime.sleep(0.05)
            os.kill(pid, signal.SIGINT)

    @run_decorator()
    def infinite_plan():
        while True:
            yield Msg("null")

    def hanging_callback(name, doc):
        for j in range(100):  # noqa: B007
            ttime.sleep(0.1)

    start_time = ttime.time()
    timer = threading.Timer(0.2, sim_kill, (11,))
    timer.start()
    with pytest.raises(RunEngineInterrupted):
        RE(infinite_plan(), {"start": hanging_callback})
    # Check that hammering SIGINT escaped from that 10-second sleep.
    assert ttime.time() - start_time < 5
    # The KeyboardInterrupt will have been converted to a hard pause.
    assert RE.state == "idle"
    timer.join()


@uses_os_kill_sigint
def test_no_context_manager(RE):
    # Clear the context managers so that RE will not react to sigint
    RE.context_managers = []

    # Proceed as normal
    pid = os.getpid()

    def sim_kill(n=1):
        for j in range(n):
            print("KILL", j)
            ttime.sleep(0.05)
            os.kill(pid, signal.SIGINT)

    def hanging_plan():
        "a plan that blocks the RunEngine's normal Ctrl+C handing with a sleep"
        ttime.sleep(2)
        yield Msg("null")

    # Only send one SIGINT
    timer = threading.Timer(0.5, sim_kill, (1,))
    timer.start()

    t = threading.Thread(target=RE, args=(hanging_plan()))
    start = ttime.time()
    t.start()

    # Wait for the KeyboardInterrupt and handle in the main thread
    with pytest.raises(KeyboardInterrupt):
        ttime.sleep(5)

    t.join()
    delta = ttime.time() - start

    # Hanging plan finished, but extra sleep did not
    assert 2 < delta < 5
    timer.join()


def test_many_context_managers(RE):
    class Manager:
        enters = 0
        exits = 0

        def __init__(self, RE):
            pass

        def __enter__(self):
            Manager.enters += 1

        def __exit__(self, *args, **kwargs):
            Manager.exits += 1

    n = 42
    RE.context_managers.extend([Manager] * n)
    RE([Msg("null")])
    assert Manager.enters == n
    assert Manager.exits == n


def _make_plan_marker():
    @reset_positions_decorator()
    def raiser(motor):
        for j in range(15):
            yield Msg("set", motor, j)
        raise RuntimeError()

    @reset_positions_decorator()
    def pausing_raiser(motor):
        for j in range(15):
            yield Msg("set", motor, j)
        yield Msg("pause")
        raise RuntimeError()

    @reset_positions_decorator()
    def bad_set(motor):
        for j in range(15):
            yield Msg("set", motor, j)
            yield Msg("set", None, j)

    @reset_positions_decorator()
    def bad_msg(motor):
        for j in range(15):
            yield Msg("set", motor, j)
        yield Msg("aardvark")

    @reset_positions_decorator()
    def cannot_pauser(motor):
        yield Msg("clear_checkpoint")
        for j in range(15):
            yield Msg("set", motor, j)
        yield Msg("pause")

    return pytest.mark.parametrize("plan", [raiser, bad_set, bad_msg, pausing_raiser, cannot_pauser])


@_make_plan_marker()
def test_cleanup_pathological_plans(RE, hw, plan):
    motor = hw.motor
    motor.set(1024)
    try:
        try:
            RE(plan(motor))
        except RunEngineInterrupted:
            pass
        if RE.state == "paused":
            assert motor.position != 1024
            RE.resume()
    except Exception:
        pass
    assert motor.position == 1024


def test_finalizer_closeable():
    pre = (j for j in range(18))
    post = (j for j in range(18))

    plan = finalize_wrapper(pre, post)

    for j in range(3):  # noqa: B007
        next(plan)
    plan.close()


def test_invalid_generator(RE, hw, capsys):
    motor = hw.motor
    from bluesky.utils import ts_msg_hook

    RE.msg_hook = ts_msg_hook

    # this is not a valid generator as it will try to yield if it
    # is throw a GeneratorExit
    def patho_finalize_wrapper(plan, post):
        try:
            yield from plan
        finally:
            yield from post

    def base_plan(motor):
        for j in range(5):
            yield Msg("set", motor, j * 2 + 1)
        yield Msg("pause")

    def post_plan(motor):
        yield Msg("set", motor, 500)

    def pre_suspend_plan():
        yield Msg("set", motor, -500)
        raise GeneratorExit("this one")

    def make_plan():
        return patho_finalize_wrapper(base_plan(motor), post_plan(motor))

    with pytest.raises(RunEngineInterrupted):
        RE(make_plan())
    RE.request_suspend(None, pre_plan=pre_suspend_plan())
    capsys.readouterr()
    try:
        RE.resume()
    except ValueError as sf:
        assert sf.__cause__.args[0] == "this one"

    actual_err, _ = capsys.readouterr()
    expected_prefix = "The plan "
    expected_postfix = (" tried to yield a value on close.  Please fix your plan.\n")[::-1]
    assert actual_err[: len(expected_prefix)] == expected_prefix
    assert actual_err[::-1][: len(expected_postfix)] == expected_postfix


def test_exception_cascade_REside(RE):
    except_hit = False

    def pausing_plan():
        nonlocal except_hit
        for j in range(5):  # noqa: B007
            yield Msg("null")
        try:
            yield Msg("pause")
        except Exception:
            except_hit = True
            raise

    def pre_plan():
        yield Msg("aardvark")

    def post_plan():
        for j in range(5):  # noqa: B007
            yield Msg("null")

    with pytest.raises(RunEngineInterrupted):
        RE(pausing_plan())
    ev = _fabricate_asycio_event(RE.loop)
    ev.set()
    RE.request_suspend(ev.wait, pre_plan=pre_plan())
    with pytest.raises(KeyError):
        RE.resume()
    assert except_hit


def test_exception_cascade_planside(RE):
    except_hit = False

    def pausing_plan():
        nonlocal except_hit
        for j in range(5):  # noqa: B007
            yield Msg("null")
        try:
            yield Msg("pause")
        except Exception:
            except_hit = True
            raise

    def pre_plan():
        yield Msg("null")
        raise RuntimeError()

    def post_plan():
        for j in range(5):  # noqa: B007
            yield Msg("null")

    with pytest.raises(RunEngineInterrupted):
        RE(pausing_plan())
    ev = _fabricate_asycio_event(RE.loop)
    ev.set()
    RE.request_suspend(ev.wait, pre_plan=pre_plan())
    with pytest.raises(RuntimeError):
        RE.resume()
    assert except_hit


def test_sideband_cancel(RE):
    loop = RE.loop
    ev = _fabricate_asycio_event(RE.loop)

    def done():
        _careful_event_set(ev)()

    def side_band_kill():
        RE.loop.call_soon_threadsafe(RE._task.cancel)

    scan = [
        Msg(
            "wait_for",
            None,
            [
                ev.wait,
            ],
        ),
    ]
    assert RE.state == "idle"
    start = ttime.time()
    timer = threading.Timer(0.5, side_band_kill)
    timer.start()
    loop.call_soon_threadsafe(loop.call_later, 2, done)
    RE(scan)
    assert RE.state == "idle"
    assert RE._task.cancelled()
    stop = ttime.time()

    assert 0.5 < (stop - start) < 2
    timer.join()


def test_no_rewind(RE):
    msg_lst = []

    def msg_collector(msg):
        msg_lst.append(msg)

    RE.rewindable = False

    plan = [Msg("null")] * 3 + [Msg("pause")] + [Msg("null")] * 3
    RE.msg_hook = msg_collector
    with pytest.raises(RunEngineInterrupted):
        RE(plan)
    RE.resume()
    assert msg_lst == plan


def test_no_rewindable_msg(RE):
    RE.rewindable = True

    msg_lst = []

    def msg_collector(msg):
        msg_lst.append(msg)

    plan = [Msg("null")] * 3 + [Msg("pause"), Msg("rewindable", None, False)] + [Msg("null")] * 3

    RE.msg_hook = msg_collector

    with pytest.raises(RunEngineInterrupted):
        RE(plan)
    RE.resume()
    assert msg_lst[:4] == plan[:4]
    assert msg_lst[4:7] == plan[:3]
    assert msg_lst[7:] == plan[4:]


@pytest.mark.parametrize("start_state", [True, False])
def test_rewindable_state_retrival(RE, start_state):
    RE.rewindable = start_state

    def rewind_plan(start_value):
        ret = yield Msg("rewindable", None, None)
        assert ret is start_state
        cache_state = ret
        ret = yield Msg("rewindable", None, start_state)
        assert ret is start_state

        ret = yield Msg("rewindable", None, not start_state)
        assert ret is (not start_state)

        ret = yield Msg("rewindable", None, cache_state)
        assert ret is start_state

    RE(rewind_plan(start_state))
    assert RE.rewindable is start_state


@pytest.mark.parametrize(
    "start_state,msg_seq",
    (
        (
            True,
            [
                "open_run",
                "declare_stream",
                "rewindable",
                "rewindable",
                "trigger",
                "trigger",
                "wait",
                "create",
                "read",
                "read",
                "save",
                "rewindable",
                "close_run",
            ],
        ),
        (
            False,
            [
                "open_run",
                "declare_stream",
                "rewindable",
                "trigger",
                "trigger",
                "wait",
                "create",
                "read",
                "read",
                "save",
                "close_run",
            ],
        ),
    ),
)
def test_nonrewindable_detector(RE, hw, start_state, msg_seq):
    class FakeSig:
        def get(self):
            return False

    hw.det.rewindable = FakeSig()

    RE.rewindable = start_state
    m_col = MsgCollector()
    RE.msg_hook = m_col

    def test():
        yield from declare_stream(hw.motor, hw.det, name="primary")
        return (yield from trigger_and_read([hw.motor, hw.det]))

    RE(run_wrapper(test()))

    assert [m.command for m in m_col.msgs] == msg_seq


@pytest.mark.parametrize(
    "start_state,msg_seq",
    ((True, ["rewindable", "rewindable", "aardvark", "rewindable"]), (False, ["rewindable", "aardvark"])),
)
def test_nonrewindable_finalizer(RE, hw, start_state, msg_seq):
    class FakeSig:
        def get(self):
            return False

    det = hw.det
    det.rewindable = FakeSig()

    RE.rewindable = start_state
    m_col = MsgCollector()
    RE.msg_hook = m_col

    def evil_plan():
        assert RE.rewindable is False
        yield Msg("aardvark")

    with pytest.raises(KeyError):
        RE(rewindable_wrapper(evil_plan(), False))

    assert RE.rewindable is start_state

    assert [m.command for m in m_col.msgs] == msg_seq


def test_halt_from_pause(RE):
    except_hit = False
    m_coll = MsgCollector()
    RE.msg_hook = m_coll

    def pausing_plan():
        nonlocal except_hit
        for j in range(5):  # noqa: B007
            yield Msg("null")
        try:
            yield Msg("pause")
        except Exception:
            yield Msg("null")
            except_hit = True
            raise

    with pytest.raises(RunEngineInterrupted):
        RE(pausing_plan())
    RE.halt()
    assert not except_hit
    assert [m.command for m in m_coll.msgs] == ["null"] * 5 + ["pause"]


def test_halt_async(RE):
    except_hit = False
    m_coll = MsgCollector()
    RE.msg_hook = m_coll

    def sleeping_plan():
        nonlocal except_hit
        try:
            yield Msg("sleep", None, 50)
        except Exception:
            yield Msg("null")
            except_hit = True
            raise

    timer = threading.Timer(0.1, RE.halt)
    timer.start()
    start = ttime.time()
    with pytest.raises(RunEngineInterrupted):
        RE(sleeping_plan())
    stop = ttime.time()
    assert 0.09 < stop - start < 5
    assert not except_hit
    assert [m.command for m in m_coll.msgs] == ["sleep"]
    timer.join()


@pytest.mark.parametrize(
    "cancel_func", [lambda RE: RE.stop(), lambda RE: RE.abort(), lambda RE: RE.request_pause(defer=False)]
)
def test_prompt_stop(RE, cancel_func):
    except_hit = False
    m_coll = MsgCollector()
    RE.msg_hook = m_coll

    def sleeping_plan():
        nonlocal except_hit
        try:
            yield Msg("sleep", None, 50)
        except Exception:
            yield Msg("null")
            except_hit = True
            raise

    timer = threading.Timer(0.1, partial(cancel_func, RE))
    timer.start()
    start = ttime.time()
    with pytest.raises(RunEngineInterrupted):
        RE(sleeping_plan())
    stop = ttime.time()
    if RE.state != "idle":
        RE.abort()
    assert 0.09 < stop - start < 5
    assert except_hit
    assert [m.command for m in m_coll.msgs] == ["sleep", "null"]
    timer.join()


@pytest.mark.parametrize(
    "change_func",
    [
        lambda RE: RE.stop(),
        lambda RE: RE.abort(),
        lambda RE: RE.halt(),
        lambda RE: RE.request_pause(),
        lambda RE: RE.request_pause(defer=True),
        lambda RE: RE.resume(),
    ],
)
def test_bad_from_idle_transitions(RE, change_func):
    with pytest.raises(TransitionError):
        change_func(RE)


def test_empty_cache_pause(RE):
    RE.rewindable = False
    pln = [
        Msg("open_run"),
        Msg("declare_stream", None, name="primary"),
        Msg("create", name="primary"),
        Msg("pause"),
        Msg("save"),
        Msg("close_run"),
    ]
    with pytest.raises(RunEngineInterrupted):
        RE(pln)
    RE.resume()


def test_state_hook(RE):
    states = []

    def log_state(new, old):
        states.append((new, old))

    RE.state_hook = log_state
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("open_run"), Msg("pause"), Msg("close_run")])
    RE.resume()
    expected = [
        ("running", "idle"),
        ("pausing", "running"),
        ("paused", "pausing"),
        ("running", "paused"),
        ("idle", "running"),
    ]
    assert states == expected


def test_max_depth(RE):
    RE.max_depth is None  # noqa: B015
    RE([])  # should not raise

    # assume test framework needs less than 100 stacks... haha
    RE.max_depth = 100
    RE([])  # should not raise

    RE.max_depth = 0
    with pytest.raises(RuntimeError):
        RE([])


def test_preprocessors(RE):
    def custom_cleanup(plan):
        yield from plan
        yield Msg("null", "cleanup")  # just a sentinel

    def my_sub(name, doc):
        pass

    def custom_subs(plan):
        yield from subs_wrapper(plan, my_sub)

    RE.preprocessors = [custom_cleanup, custom_subs]
    actual = []
    RE.msg_hook = lambda msg: actual.append(msg)
    RE([Msg("null")])
    print(actual)
    expected = [
        Msg("subscribe", None, my_sub, "all"),
        Msg("null"),
        Msg("null", "cleanup"),
        Msg("unsubscribe", None, token=0),
    ]
    assert actual == expected


@requires_ophyd
def test_pardon_failures(RE):
    from ophyd import StatusBase

    st = StatusBase()

    class Dummy:
        name = "dummy"

        def set(self, val):
            return st

    dummy = Dummy()

    RE([Msg("set", dummy, 1)])
    st._finished(success=False)
    RE([Msg("null")])


@requires_ophyd
def test_exceptions_kill_run_and_failed_status_holds_detail(RE):
    # just to make sure that 'pardon_failures' does not block *real* failures
    from ophyd import StatusBase
    from ophyd.utils.errors import UnknownStatusFailure

    class Dummy:
        name = "dummy"

        def set(self, val):
            st = StatusBase()
            st._finished(success=False)
            return st

    dummy = Dummy()

    with pytest.raises(FailedStatus) as exc:
        RE([Msg("set", dummy, 1, group="test"), Msg("wait", group="test")])
        assert isinstance(exc.args[0], UnknownStatusFailure)


@requires_ophyd
def test_wait_with_timeout_can_be_repeated(RE):
    from ophyd import StatusBase

    class Dummy:
        name = "dummy"

        def set(self, val):
            self.st = StatusBase()
            return self.st

    def plan():
        dummy = Dummy()
        yield Msg("set", dummy, 1, group="test")
        with pytest.raises(TimeoutError):
            yield Msg("wait", group="test", timeout=0.1)
        with pytest.raises(TimeoutError):
            yield Msg("wait", group="test", timeout=0.1)
        dummy.st.set_finished()
        yield Msg("wait", group="test")

    RE(plan())


@requires_ophyd
def test_status_propagates_exception_through_run_engine(RE: RunEngine):
    # just to make sure that 'pardon_failures' does not block *real* failures
    from ophyd import StatusBase

    class DummyV1:
        name = "dummy"

        def set(self, val):
            st = StatusBase()
            try:
                1 / 0  # noqa: B018
            except ZeroDivisionError as exc:
                st.set_exception(exc)

            return st

    dummy1 = DummyV1()

    with pytest.raises(FailedStatus) as exc:
        RE([Msg("set", dummy1, 1, group="test"), Msg("wait", group="test")])

        traceback: list[FrameSummary] = extract_tb(exc.__traceback__)  # type: ignore[attr-defined]
        assert traceback[0].filename == __file__
        assert traceback[0].line == "RE([Msg('set', dummy1, 1, group='test'),"
        assert traceback[-1].name == "set"
        assert traceback[-1].line == "1/0"

        assert isinstance(exc.args[0], ZeroDivisionError)  # type: ignore[attr-defined]


def test_colliding_streams(RE, hw):
    collector = {"primary": [], "baseline": []}
    descs = {}

    def local_cb(name, doc):
        if name == "descriptor":
            descs[doc["uid"]] = doc["name"]
        elif name == "event":
            collector[descs[doc["descriptor"]]].append(doc)

    RE(
        baseline_wrapper(
            grid_scan([hw.motor], hw.motor, -1, 1, 5, hw.motor1, -5, 5, 7, True), [hw.motor, hw.motor1]
        ),
        local_cb,
    )

    assert len(collector["primary"]) == 35
    assert len(collector["baseline"]) == 2

    assert list(range(1, 36)) == [e["seq_num"] for e in collector["primary"]]
    assert list(range(1, 3)) == [e["seq_num"] for e in collector["baseline"]]


def test_old_subscribe(RE):
    # Old usage had reversed argument order. It should warn but still work.
    collector = []

    def collect(name, doc):
        collector.append(doc)

    with pytest.warns(UserWarning):
        RE.subscribe("all", collect)

    RE([Msg("open_run"), Msg("close_run")])
    assert len(collector) == 2

    RE.unsubscribe(0)
    with pytest.warns(UserWarning):
        RE.subscribe("start", collect)

    RE([Msg("open_run"), Msg("close_run")])
    assert len(collector) == 3

    RE.unsubscribe(1)


def test_waiting_hook(RE, hw):
    collector = []

    def collect(sts):
        collector.append(sts)

    RE.waiting_hook = collect

    RE([Msg("set", hw.motor, 5, group="A"), Msg("wait", group="A")])

    sts, none = collector
    assert isinstance(sts, set)
    assert len(sts) == 1
    assert none is None
    collector.clear()

    RE([Msg("set", hw.motor1, 5, group="A"), Msg("set", hw.motor2, 3, group="A"), Msg("wait", group="A")])

    sts, none = collector
    assert isinstance(sts, set)
    assert len(sts) == 2
    assert none is None
    collector.clear()


def test_hints(RE):
    class Detector:
        def __init__(self, name):
            self.name = name
            self.parent = None
            self.hints = {"vis": "placeholder"}

        async def read(self):
            await asyncio.sleep(0.1)
            return {}

        async def describe(self):
            return {}

        async def read_configuration(self):
            return {}

        async def describe_configuration(self):
            return {}

    det = Detector("det")

    collector = []

    RE(count([det]), {"descriptor": lambda name, doc: collector.append(doc)})
    doc = collector.pop()
    assert doc["hints"]["det"] == {"vis": "placeholder"}


def test_filled(RE, hw, db):
    collector = []

    def collect(name, doc):
        if name == "event":
            collector.append(doc)

    RE(count([hw.det]), collect)

    (event,) = collector
    assert event["filled"] == {}
    collector.clear()

    RE(count([hw.img]), collect)
    (event,) = collector
    assert event["filled"] == {"img": False}


def test_double_call(RE):
    uid1 = RE(count([]))
    uid2 = RE(count([]))

    assert uid1 != uid2


@pytest.mark.parametrize("predeclare", [True, False])
def test_num_events(RE, hw, db, predeclare, monkeypatch):
    if predeclare:
        monkeypatch.setenv("BLUESKY_PREDECLARE", "1")
    RE.subscribe(db.insert)

    rs1 = RE(count([]))
    if RE.call_returns_result:
        uid1 = rs1.run_start_uids[0]
    else:
        uid1 = rs1[0]
    h = db[uid1]
    if predeclare:
        assert h.stop["num_events"] == {"primary": 0}
    else:
        assert h.stop["num_events"] == {}

    rs2 = RE(count([hw.det], 5))
    if RE.call_returns_result:
        uid2 = rs2.run_start_uids[0]
    else:
        uid2 = rs2[0]
    h = db[uid2]
    assert h.stop["num_events"] == {"primary": 5}

    sd = SupplementalData(baseline=[hw.det])
    RE.preprocessors.append(sd)

    rs3 = RE(count([]))
    if RE.call_returns_result:
        uid3 = rs3.run_start_uids[0]
    else:
        uid3 = rs3[0]
    h = db[uid3]
    if predeclare:
        assert h.stop["num_events"] == {"primary": 0, "baseline": 2}
    else:
        assert h.stop["num_events"] == {"baseline": 2}

    rs4 = RE(count([hw.det], 5))
    if RE.call_returns_result:
        uid4 = rs4.run_start_uids[0]
    else:
        uid4 = rs4[0]
    h = db[uid4]
    assert h.stop["num_events"] == {"primary": 5, "baseline": 2}


def test_raise_if_interrupted_deprecation(RE):
    with pytest.warns(UserWarning):
        RE([], raise_if_interrupted=True)


@pytest.mark.parametrize(
    "bail_func,status", (("resume", "success"), ("stop", "success"), ("abort", "abort"), ("halt", "abort"))
)
def test_force_stop_exit_status(bail_func, status, RE):
    d = DocCollector()
    RE.subscribe(d.insert)

    @run_decorator()
    def bad_plan():
        print("going in")
        yield Msg("pause")

    with pytest.raises(RunEngineInterrupted):
        RE(bad_plan())

    rs = getattr(RE, bail_func)()

    if RE.call_returns_result:
        if bail_func == "resume":
            assert rs.plan_result is not RE.NO_PLAN_RETURN
        else:
            assert rs.plan_result is RE.NO_PLAN_RETURN
        uid = rs.run_start_uids[0]
        assert rs.exit_status == status
    else:
        uid = rs[0]
    assert len(d.start) == 1
    assert d.start[0]["uid"] == uid
    assert len(d.stop) == 1
    assert d.stop[uid]["exit_status"] == status


def test_exceptions_exit_status(RE):
    d = DocCollector()
    RE.subscribe(d.insert)

    class Snowflake(Exception): ...

    @run_decorator()
    def bad_plan():
        yield Msg("null")
        raise Snowflake("boo")

    with pytest.raises(Snowflake) as sf:
        RE(bad_plan())

    assert len(d.start) == 1
    rs = d.start[0]["uid"]
    assert len(d.stop) == 1
    assert d.stop[rs]["exit_status"] == "fail"
    assert d.stop[rs]["reason"] == str(sf.value)


def test_plan_return(RE):
    if not RE.call_returns_result:
        pytest.skip()

    def test_plan():
        yield Msg("null")
        return "plan_result"

    rs = RE(test_plan())
    assert rs.plan_result == "plan_result"
    assert rs.exit_status == "success"
    assert rs.interrupted is False


def test_plan_return_resume(RE):
    if not RE.call_returns_result:
        pytest.skip()

    def test_plan():
        yield Msg("null")
        yield Msg("pause")
        return "plan_result"

    with pytest.raises(RunEngineInterrupted):
        RE(test_plan())
    rs = RE.resume()
    assert rs.plan_result == "plan_result"
    assert rs.exit_status == "success"


def test_drop(RE, hw):
    det = hw.det

    def inner(msg):
        yield Msg("create", name="primary")
        yield Msg("read", det)
        yield Msg(msg)

    # Drop first, drop after saving, save after dropping
    def plan():
        yield Msg("open_run")
        yield Msg("declare_stream", None, det, name="primary")
        yield from inner("drop")
        yield from inner("save")
        yield from inner("drop")
        yield from inner("save")
        yield from inner("save")
        yield Msg("close_run")

    docs = defaultdict(list)

    def collector(name, doc):
        docs[name].append(doc)

    RE(plan(), collector)

    assert len(docs["event"]) == 3


def test_failing_describe_callback(RE, hw):
    class TestException(Exception):
        pass

    det = hw.det
    det2 = hw.det2

    def evil_cb(name, doc):
        if any(k in doc["data_keys"] for k in det.describe()):
            raise TestException

    RE.subscribe(evil_cb, "descriptor")

    def plan():
        yield Msg("open_run")
        try:
            yield Msg("declare_stream", None, det, name="det1")
            yield Msg("create", name="det1")
            yield Msg("read", det)
            yield Msg("save")
        finally:
            yield Msg("declare_stream", None, det2, name="det2")
            yield Msg("create", name="det2")
            yield Msg("read", det2)
            yield Msg("save")
        yield Msg("close_run")

    with pytest.raises(TestException):
        RE(plan())


def test_print_commands(RE):
    """test the printing of commands available.
    NOTE : An error here likely just means that this interface has changed.
    You'll likely want to change the test.
    (A test here is still a good idea as it at least raises awareness about
    the changes made breaking past API)
    """

    # testing the commands list
    commands1 = list(RE._command_registry.keys())
    commands2 = RE.commands

    assert commands1 == commands2

    # testing print commands
    # copy and paste most of the code...
    verbose = False
    print_command_reg1 = "List of available commands\n"
    for command, func in RE._command_registry.items():
        docstring = func.__doc__
        if verbose is False:
            docstring = docstring.split("\n")[0]
        print_command_reg1 += f"{command} : {docstring}\n"

    print_command_reg2 = RE.print_command_registry()
    assert print_command_reg1 == print_command_reg2


def test_broken_read_exception(RE):
    class Dummy:
        def __init__(self, name):
            self.name = name

        def read_configuration(self):
            return {}

        def describe_configuration(self):
            return {}

        def describe(self):
            return {}

        def read(self): ...

        def trigger(self): ...

    obj = Dummy("broken read")
    with pytest.raises(RuntimeError):
        RE([Msg("read", obj)])


def test_self_describe(RE):
    def inner():
        cls = yield Msg("RE_class")
        assert isinstance(RE, cls)

    RE(inner())


def test_thread_name(RE):
    "The RunEngine event loop should be on a thread with a given name."
    from ophyd.status import Status

    class MockDevice:
        name = "mock_device"

        def trigger(self):
            assert threading.current_thread().name == "bluesky-run-engine"
            return Status()

        def describe(self):
            return {}

        def read(self):
            return {}

    d = MockDevice()
    RE([Msg("trigger", d)])


def test_unsubscribe(RE):
    def foo(name, doc): ...

    for j in range(15):  # noqa: B007
        cid = RE.subscribe(foo)
        RE.unsubscribe(cid)

    assert len(RE.dispatcher._token_mapping) == 0


@pytest.fixture
def device_pair():
    from ophyd import StatusBase
    from ophyd.device import Device

    class MockDevice(Device):
        def stage(self):
            return StatusBase()

    return MockDevice(name="device1"), MockDevice(name="device2")


@pytest.mark.parametrize("set_finished", [True, False])
def test_wait_with_timeout(set_finished, RE, device_pair):
    def plan():
        status = yield Msg("stage", device_pair[0], group="test_group")
        if set_finished:
            status.set_finished()
        yield from wait(group="test_group", timeout=0.1)

    if set_finished:
        RE(plan())
    else:
        with pytest.raises(TimeoutError):
            RE(plan())


def test_wait_then_add_to_groups(RE, device_pair):
    def plan():
        status1 = yield Msg("stage", device_pair[0], group="test_group")
        done = yield from wait(group="test_group", timeout=0.001, error_on_timeout=False)
        assert done is False
        status2 = yield Msg("stage", device_pair[1], group="test_group")
        done = yield from wait(group="test_group", timeout=0.001, error_on_timeout=False)
        assert done is False
        status1.set_finished()
        status2.set_finished()
        done = yield from wait(group="test_group", timeout=0.001)
        assert done is True

    RE(plan())


async def finish_status_later(status, when: float, error: bool):
    await asyncio.sleep(when)
    if error:
        status.set_exception(ValueError("failed"))
    else:
        status.set_finished()


@pytest.mark.parametrize(
    "raises,exception,duration",
    [
        (True, FailedStatus, 0.1),
        (False, WaitForTimeoutError, 0.5),
    ],
)
def test_wait_when_one_status_returns_raises(RE, device_pair, raises, exception, duration):
    def plan():
        yield Msg("stage", device_pair[0], group="test_group")
        status2 = yield Msg("stage", device_pair[1], group="test_group")
        asyncio.create_task(finish_status_later(status2, 0.05, raises))
        yield from wait(group="test_group", timeout=0.5)

    start = ttime.monotonic()
    with pytest.raises(exception):
        RE(plan())
    elapsed = ttime.monotonic() - start
    assert elapsed == pytest.approx(duration, abs=0.1)


@pytest.mark.parametrize(
    "raises,exception,duration",
    [
        (True, FailedStatus, 0.1),
        (False, WaitForTimeoutError, 0.5),
    ],
)
def test_wait_returns_based_on_watch_raises(RE, device_pair, raises, exception, duration):
    def plan():
        yield Msg("stage", device_pair[0], group="test_group")
        stage2 = yield Msg("stage", device_pair[1], group="watch_group")
        asyncio.create_task(finish_status_later(stage2, 0.05, raises))
        yield from wait(group="test_group", watch=("watch_group",), timeout=0.5)

    start = ttime.monotonic()
    with pytest.raises(exception):
        RE(plan())
    elapsed = ttime.monotonic() - start
    assert elapsed == pytest.approx(duration, abs=0.1)


def test_set_finished_before_watch_completes_immediately(RE, device_pair):
    def plan():
        status = yield Msg("stage", device_pair[0], group="test_group")
        status.set_finished()
        yield Msg("stage", device_pair[1], group="watch_group")
        yield from wait(group="test_group", timeout=0.1, watch=("watch_group",))

    RE(plan())


def test_watch_finished_before_set_return_when_set_finishes(RE, device_pair):
    def plan():
        status1 = yield Msg("stage", device_pair[0], group="test_group")
        status2 = yield Msg("stage", device_pair[1], group="watch_group")
        asyncio.create_task(finish_status_later(status1, 0.2, False))
        asyncio.create_task(finish_status_later(status2, 0.05, False))
        yield from wait(group="test_group", timeout=0.3, watch=("watch_group",))

    start = ttime.monotonic()
    RE(plan())
    elapsed = ttime.monotonic() - start
    assert elapsed == pytest.approx(0.2, abs=0.05)


def test_set_failed_before_watch_raises_immediately(RE, device_pair):
    def plan():
        status = yield Msg("stage", device_pair[0], group="test_group")
        yield Msg("stage", device_pair[1], group="watch_group")
        status.set_exception(Exception("Mock device failed"))
        yield from wait(group="test_group", timeout=1.0, watch=("watch_group",))

    start = ttime.monotonic()
    with pytest.raises(FailedStatus):
        RE(plan())
    elapsed = ttime.monotonic() - start
    assert elapsed == pytest.approx(0.0, abs=0.1)


async def passing_coroutine():
    await asyncio.sleep(0.001)
    return 1080


async def failing_coroutine():
    await asyncio.sleep(0.001)
    raise RuntimeError("This is a runtime error")


def test_wait_with_planstub_coroutine(RE):
    def runtime_error_and_success_plan():
        tasks = yield from wait_for([lambda: passing_coroutine(), lambda: failing_coroutine()])
        passing_future, failing_future = tasks
        assert isinstance(passing_future, asyncio.Task) and isinstance(failing_future, asyncio.Task)
        assert passing_future.get_coro().__name__ == "passing_coroutine"
        assert failing_future.get_coro().__name__ == "failing_coroutine"
        assert passing_future.done() and passing_future.result() == 1080
        assert failing_future.done() and isinstance(failing_future.exception(), RuntimeError)

        with pytest.raises(RuntimeError, match="This is a runtime error"):
            raise failing_future.exception()  # the error is the same

    RE(runtime_error_and_success_plan())


def test_wait_with_planstub_status(RE):
    def runtime_error_and_success_plan():
        passing_task = asyncio.create_task(passing_coroutine())

        failing_task = asyncio.create_task(failing_coroutine())

        futures = yield from wait_for([lambda: passing_task, lambda: failing_task])
        assert passing_task is futures[0] and failing_task is futures[1]
        assert passing_task.done() and passing_task.result() == 1080
        assert failing_task.done() and isinstance(failing_task.exception(), RuntimeError)

    RE(runtime_error_and_success_plan())


def test_wait_with_planstub_timeout(RE):
    def timeout_plan():
        async def timeout_coro():
            await asyncio.sleep(1e9)

        timeout_task = asyncio.create_task(timeout_coro())
        try:
            yield from wait_for([lambda: timeout_task], timeout=0.01)
        finally:
            # Manually tear down the tasks, otherwise it can happen
            # after the test succeeds/fails
            timeout_task.cancel()

    with pytest.raises(WaitForTimeoutError, match="Plan failed to complete in the specified time"):
        RE(timeout_plan())


def test_1event_rewind(RE, hw):
    cache = []

    def hook(msg):
        cache.append(msg)

    doc_cache = []

    def cb(name, doc):
        doc_cache.append((name, doc))

    @run_decorator()
    def broken():
        yield from checkpoint()
        yield from trigger_and_read([hw.det], name="primary")
        yield from pause()
        yield from trigger_and_read([hw.det], name="primary")

    RE.msg_hook = hook
    with pytest.raises(RunEngineInterrupted):
        RE(broken(), cb)
    RE.resume()

    assert [_[1]["seq_num"] for _ in doc_cache if _[0] == "event"] == [1, 1, 2]

    assert [_[0] for _ in doc_cache] == [
        "start",
        "descriptor",
        "event",
        "event",
        "event",
        "stop",
    ]

    assert [msg.command for msg in cache] == [
        "open_run",
        "checkpoint",
        "trigger",
        "wait",
        "create",
        "read",
        "save",
        "pause",
        "trigger",
        "wait",
        "create",
        "read",
        "save",
        "trigger",
        "wait",
        "create",
        "read",
        "save",
        "close_run",
    ]


@requires_ophyd
def test_configure_multiple_descritpors(RE):
    from ophyd import Component as C
    from ophyd import Device, sim

    class SynWithConfig(Device):
        x = C(sim.Signal, value=0)
        y = C(sim.Signal, value=2)
        z = C(sim.Signal, value=3)

    det = SynWithConfig(name="det")
    det.x.name = "x"
    det.y.name = "y"
    det.z.name = "z"
    det.read_attrs = ["x"]
    det.configuration_attrs = ["y", "z"]

    @run_decorator()
    def plan(det):
        # run without an exciting descriptor
        yield from configure(det, {"z": 3})
        # force a descriptor to be created
        yield from declare_stream(det, name="primary")
        # and an event
        yield from trigger_and_read([det], name="primary")
        # update the config which will re-generate the descriptor
        yield from configure(det, {"z": 4})
        # and a second event
        yield from trigger_and_read([det], name="primary")

    d = DocCollector()
    RE(plan(det), d.insert)

    (start,) = d.start
    descA, descB = d.descriptor[start["uid"]]
    stop = d.stop[start["uid"]]

    assert descA["configuration"]["det"]["data"]["z"] == 3
    assert descB["configuration"]["det"]["data"]["z"] == 4

    for desc in (descA, descB):
        assert len(d.event[desc["uid"]]) == 1

    assert stop["num_events"]["primary"] == 2


def test_sync_scan_id_source(RE):
    def sync_scan_source(md: dict) -> int:
        return 314159

    RE.scan_id_source = sync_scan_source
    RE([Msg("open_run")])
    assert RE.md["scan_id"] == 314159


def test_async_scan_id_source(RE):
    async def async_scan_source(md: dict) -> int:
        return 42

    RE.scan_id_source = async_scan_source
    RE([Msg("open_run")])
    assert RE.md["scan_id"] == 42


@requires_ophyd
def test_descriptor_order(RE):
    from itertools import permutations

    from ophyd import Component, Device, Signal

    class Issue1930(Device):
        alpha = Component(Signal, value=1, kind="hinted")
        bravo = Component(Signal, value=2, kind="hinted")
        charlie = Component(Signal, value=3, kind="hinted")

    i1930 = Issue1930(name="i1930")

    for dets in permutations([i1930.alpha, i1930.bravo, i1930.charlie]):
        key_order = [d.name for d in dets]

        def check(key_order, name, doc):
            if name == "event":
                assert list(doc["data"]) == key_order
            elif name == "descriptor":
                assert list(doc["data_keys"]) == key_order

        RE(scan(dets, i1930.charlie, -1, 1, 2), lambda name, doc, key_order=key_order: check(key_order, name, doc))



import hashlib
import sys
import time
from functools import partial
from threading import RLock
from typing import Any, Optional, TextIO

import numpy as np
from IPython.core.display import HTML
from IPython.display import display
from ipywidgets import HBox
from tqdm.notebook import tqdm as tqdm_nb

from bluesky.utils import ProgressBarBase, ProgressBarManager, _L2norm


class NotebookProgressBar(ProgressBarBase):
    containers: list[HBox]
    status_objs: list[Any]
    fp: TextIO
    creation_time: float
    lock: RLock

    def __init__(self, status_objs: list[Any], delay_draw: float = 0.2):
        """
        Represents status objects with Jupyter Notebook progress bars.

        Parameters
        ----------
        status_objs : list
            Status objects
        """
        self.containers = []
        self.status_objs = []
        self.fp = sys.stdout
        self.creation_time = time.time()
        self.delay_draw = delay_draw
        self.lock = RLock()

        # Create a closure over self.update for each status object that
        # implemets the 'watch' method.
        for st in status_objs:
            with self.lock:
                if hasattr(st, "watch") and not st.done:
                    pos = len(self.containers)
                    container: HBox = tqdm_nb.status_printer(self.fp, 1, None, None)
                    display(container)
                    self.containers.append(container)
                    self.status_objs.append(st)
                    st.watch(partial(self.update, pos))

    def update(
        self,
        pos: Any,
        *,
        name: Optional[str] = None,
        current: Any = None,
        initial: Any = None,
        target: Any = None,
        unit: str = "units",
        precision: Any = None,
        fraction: Any = None,
        time_elapsed: Optional[float] = None,
        time_remaining: Optional[float] = None,
    ):
        if all(x is not None for x in (current, initial, target)):
            # In this case there is enough information to draw a progress bar with
            # current and maximum values
            total = round(_L2norm(target, initial), precision or 3)
            n = np.clip(round(_L2norm(current, initial), precision or 3), 0, total)
            if time_elapsed is None:
                time_elapsed = time.time() - self.creation_time
            meta = tqdm_nb.format_meter(
                n=n,
                total=total,
                elapsed=time_elapsed,
                unit=unit,
                prefix=name,
                ncols=79,
                bar_format="{r_bar}",
            )
            self.draw(
                pos,
                name or "",
                meta,
                color=html_color_from_name(name or ""),
                total=total,
                value=n,
            )
        elif self.status_objs[pos].done:
            # In this case the operation is complete, we just
            # display a finished progress bar
            # and a completion message.
            self.draw(
                pos,
                name or "",
                "[Complete.]",
                color=html_color_from_name(name or ""),
                value=1,
            )
        else:
            # In this case there is not enough information to draw a progress bar so
            # we just display a message.
            self.draw(pos, name or "", "[In progress. No progress bar available.]")

    def draw(
        self,
        pos: int,
        label: str = "",
        meta: str = "",
        color: str = "#97d4e8",
        total: float = 1.0,
        value: Optional[float] = None,
    ) -> None:
        """
        Draws the progress bar or a message if there is
        not enough information

        Parameters
        ----------
        pos : int
            The index of the progress bar in self.containers
        label : str, optional
            The label to go on the left-hand side of the progress bar, by default ""
        meta : str, optional
            Any metadata to go on the right-hand side of the progress bar, by default ""
        color : str, optional
            The color of the progress bar as a valid HTML/CSS string, by default "#97d4e8"
        total : float, optional
            The maximum value of the progress bar, defaults to 1, by default 1.0
        value : Optional[float], optional
            The current value of the progress bar, optional, by default None
        """

        # Do not draw if this was only created recently, otherwise it
        # will just flash up and disappear
        if (time.time() - self.creation_time) < self.delay_draw:
            return

        container = self.containers[pos]
        with self.lock:
            ltext, pbar, rtext = container.children
            ltext.value = label
            pbar.layout.display = "flex"
            rtext.value = meta

            if value is not None:
                pbar.value = value
                pbar.max = total or 1
                pbar.style.bar_color = color
            else:
                pbar.layout.display = "none"

    def clear(self) -> None:
        """
        Clears all progress bars
        """

        with self.lock:
            for container in self.containers:
                container.close()
        hide_leftover_progress_bars()


def html_color_from_name(name: str) -> str:
    """
    Transforms a string into a unique, bright HTML color
    (hue varies, saturation and value are fixed).
    Transformation is pure.

    Parameters
    ----------
    name : str
        The string to convert

    Returns
    -------
    str
        A string describing a color, valid in HTML/CSS.
    """

    hue = int(hashlib.sha1(name.encode("utf-8")).hexdigest(), 16) % 360
    return f"hsl({hue}, 90%, 55%)"


def hide_leftover_progress_bars() -> None:
    """
    Sets the style in Jupyterlab to remove leftover whitespace
    from deleted progress bars. Workaround for:
    https://github.com/jupyterlab/jupyterlab/issues/7354
    """

    display(
        HTML("""
            <style>
                .p-Widget.jp-OutputPrompt.jp-OutputArea-prompt:empty {
                    padding: 0;
                    border: 0;
                }
            </style>
        """)
    )


def pbar_manager_for_notebook(delay_draw: float = 0.2) -> ProgressBarManager:
    """
    Helper method for generating managers when using notebooks.

    Returns
    -------
    ProgressBarManager
        A manager that creates progress bars for Jupuyter notebooks
    """

    return ProgressBarManager(partial(NotebookProgressBar, delay_draw=delay_draw))



import re
from collections.abc import Iterator
from typing import Optional

import pytest
from event_model.documents import Datum
from event_model.documents.event import PartialEvent
from event_model.documents.event_descriptor import DataKey
from event_model.documents.event_page import PartialEventPage
from event_model.documents.resource import PartialResource
from event_model.documents.stream_datum import StreamDatum
from event_model.documents.stream_resource import StreamResource

import bluesky.plan_stubs as bps
import bluesky.plans as bp
from bluesky.protocols import (
    Asset,
    Collectable,
    EventCollectable,
    EventPageCollectable,
    HasName,
    Readable,
    Reading,
    StreamAsset,
    WritesExternalAssets,
    WritesStreamAssets,
)


class DocHolder(dict):
    def append(self, name, doc):
        self.setdefault(name, []).append(doc)

    def assert_emitted(self, **numbers: int):
        assert list(self) == list(numbers)
        assert {name: len(d) for name, d in self.items()} == numbers


def merge_dicts(*functions):
    def wrapper(self):
        ret = {}
        for f in functions:
            ret.update(f(self))
        return ret

    return wrapper


def merge_iterators(*iterators):
    def wrapper(self):
        for it in iterators:
            yield from it(self)

    return wrapper


def collect_plan(*objs, pre_declare: bool, stream=False, stream_name=None):
    yield from bps.open_run()
    if pre_declare:
        yield from bps.declare_stream(*objs, name=stream_name, collect=True)
    yield from bps.collect(*objs, stream=stream, name=stream_name)
    yield from bps.close_run()


def collect_plan_no_stream_name_in_collect(*objs, pre_declare: bool, stream=False, stream_name=None):
    yield from bps.open_run()
    if pre_declare:
        yield from bps.declare_stream(*objs, name=stream_name, collect=True)
    yield from bps.collect(*objs, stream=stream)
    yield from bps.close_run()


class Named(HasName):
    name: str = ""
    parent = None

    def __init__(self, name: str) -> None:
        self.name = name
        self.counter = 0


def describe_datum(self: Named) -> dict[str, DataKey]:
    """Describe a single data key backed with Resource"""
    return {f"{self.name}-datum": DataKey(source="file", dtype="number", shape=[1000, 1500], external="OLD:")}


def collect_asset_docs_datum(self) -> Iterator[Asset]:
    """Produce Resource and Datum for a single frame in a file"""
    resource = PartialResource(
        resource_kwargs={"resource_arg": 42},
        root="/root",
        spec="ADHDF5_SWMR",
        resource_path="/path.h5",
        uid="RESOURCEUID",
    )
    yield "resource", resource
    datum = Datum(
        datum_id="RESOURCEUID/1",
        resource="RESOURCEUID",
        datum_kwargs={"datum_arg": 43},
    )
    yield "datum", datum


def read_datum(self: Named) -> dict[str, Reading]:
    """Read a single reference to a Datum"""
    return {f"{self.name}-datum": {"value": "RESOURCEUID/1", "timestamp": 456}}


class DatumReadable(Named, Readable, WritesExternalAssets):
    """A readable that produces a single Event backed by a Datum"""

    describe = describe_datum
    read = read_datum
    collect_asset_docs = collect_asset_docs_datum


def describe_stream_datum(self: Named) -> dict[str, DataKey]:
    """Describe 2 datasets which will be backed by StreamResources"""
    return {
        f"{self.name}-sd1": DataKey(source="file", dtype="number", shape=[1000, 1500], external="STREAM:"),
        f"{self.name}-sd2": DataKey(source="file", dtype="number", shape=[], external="STREAM:"),
    }


def get_index(self) -> int:
    """Report how many frames were written"""
    return 10


def collect_asset_docs_stream_datum(self: Named, index: Optional[int] = None) -> Iterator[StreamAsset]:
    """Produce a StreamResource and StreamDatum for 2 data keys for 0:index"""
    index = index or 1
    for data_key in [f"{self.name}-sd1", f"{self.name}-sd2"]:
        uid = f"{data_key}-uid"
        if self.counter == 0:
            # Backward compatibility test, ignore typing errors
            stream_resource = StreamResource(  # type: ignore[typeddict-item]
                resource_kwargs={"dataset": f"/{data_key}/data"},
                data_key=data_key,
                root="/root",
                resource_path="/path.h5",
                spec="AD_HDF5_SWMR_STREAM",
                uid=uid,
            )
            yield "stream_resource", stream_resource

        stream_datum = StreamDatum(
            stream_resource=uid,
            descriptor="",
            uid=f"{uid}/{self.counter}",
            indices={"start": self.counter, "stop": self.counter + index},
            seq_nums={"start": 0, "stop": 0},
        )
        yield "stream_datum", stream_datum
    self.counter += index


def read_empty(self) -> dict[str, Reading]:
    """Produce an empty event"""
    return {}


def describe_pv(self: Named) -> dict[str, DataKey]:
    """Describe a single data_key backed by a PV value"""
    return {f"{self.name}-pv": DataKey(source="pv", dtype="number", shape=[])}


def read_pv(self: Named) -> dict[str, Reading]:
    """Read a single data_key from a PV"""
    return {f"{self.name}-pv": Reading(value=5.8, timestamp=123)}


class PvAndDatumReadable(Named, Readable, WritesExternalAssets):
    """An odd device that produces a single event with one datakey from a pv, and one backed by a Datum"""

    describe = merge_dicts(describe_pv, describe_datum)
    read = merge_dicts(read_pv, read_datum)
    collect_asset_docs = collect_asset_docs_datum


class PvAndStreamDatumReadable(Named, Readable, WritesStreamAssets):
    """An odd device that produces a single event with one datakey from a pv, and one backed by a StreamDatum"""

    describe = merge_dicts(describe_pv, describe_stream_datum)
    read = read_pv
    collect_asset_docs = collect_asset_docs_stream_datum
    get_index = get_index


def describe_collect_old_pv(self) -> dict[str, dict[str, DataKey]]:
    """Doubly nested old describe format with 2 pvs in each stream"""
    return {
        stream: {f"{stream}-{pv}": DataKey(source="pv", dtype="number", shape=[]) for pv in ["pv1", "pv2"]}
        for stream in ["stream1", "stream2"]
    }


def collect_old_pv(self) -> Iterator[PartialEvent]:
    """Produce 2 events in each of the 2 streams"""
    for i in range(2):
        for stream in ["stream1", "stream2"]:
            yield PartialEvent(
                data={f"{stream}-{pv}": i for pv in ["pv1", "pv2"]},
                timestamps={f"{stream}-{pv}": 100 + i for pv in ["pv1", "pv2"]},
                time=100 + i,
            )


class OldPvCollectable(Named, EventCollectable):
    """Produce events in 2 streams backed by PVs"""

    describe_collect = describe_collect_old_pv
    collect = collect_old_pv


def describe_collect_old_datum(self: Named):
    """Produces a single data key in a stream backed by a Datum"""
    return {
        "stream3": {
            f"{self.name}-datum": DataKey(source="file", dtype="number", shape=[1000, 1500], external="OLD:")
        }
    }


def collect_old_datum(self):
    """Produces a single event backed by a Datum"""
    yield PartialEvent(
        data={f"{self.name}-datum": "RESOURCEUID/1"},
        timestamps={f"{self.name}-datum": 456},
        filled={f"{self.name}-datum": False},
    )


class OldDatumCollectable(Named, EventCollectable, WritesExternalAssets):
    """Produces a single stream with a single event in it backed by a Datum"""

    describe_collect = describe_collect_old_datum
    collect = collect_old_datum
    collect_asset_docs = collect_asset_docs_datum


def describe_collect_two_streams_same_names(self):
    """Buggy old describe output with the same data key in 2 streams"""
    return {stream: {"pv": DataKey(source="pv", dtype="number", shape=[])} for stream in ["stream1", "stream2"]}


def collect_one_pv(self):
    """Event producing value of a single pv"""
    yield PartialEvent(data={"pv": 0}, timestamps={"pv": 100})


class MultiKeyOldCollectable(Named, EventCollectable):
    """Buggy Device that should fail describe_collect as it makes 2 streams with the same data_key"""

    describe_collect = describe_collect_two_streams_same_names
    collect = collect_one_pv


class OldPvAndDatumCollectable(Named, EventCollectable, WritesExternalAssets):
    """Old style Device that produces 2 streams with PVs and 1 stream with Datum"""

    describe_collect = merge_dicts(describe_collect_old_pv, describe_collect_old_datum)
    collect = merge_iterators(collect_old_pv, collect_old_datum)
    collect_asset_docs = collect_asset_docs_datum


def describe_collect_pv(self: Named) -> dict[str, DataKey]:
    """New style describe collect with 2 PVs"""
    return {f"{self.name}-{pv}": DataKey(source="pv", dtype="number", shape=[]) for pv in ["pv1", "pv2"]}


def collect_pv(self: Named) -> Iterator[PartialEvent]:
    """Collect for the new style describe collect of 2 data keys from PVs"""
    for i in range(2):
        yield PartialEvent(
            data={f"{self.name}-{pv}": i for pv in ["pv1", "pv2"]},
            timestamps={f"{self.name}-{pv}": 100 + i for pv in ["pv1", "pv2"]},
            time=100 + i,
        )


def collect_pages_pv(self: Named) -> Iterator[PartialEventPage]:
    """Same as collect_pv but in EventPage form"""
    yield PartialEventPage(
        data={f"{self.name}-{pv}": [0, 1] for pv in ["pv1", "pv2"]},
        timestamps={f"{self.name}-{pv}": [100, 101] for pv in ["pv1", "pv2"]},
        time=[100, 101],
    )


class PvCollectable(Named, EventCollectable):
    """Produces events with 2 data keys backed by PVs"""

    describe_collect = describe_collect_pv
    collect = collect_pv


class PvPageCollectable(Named, EventPageCollectable):
    """Produces event pages with 2 data keys backed by PVs"""

    describe_collect = describe_collect_pv
    collect_pages = collect_pages_pv


class StreamDatumReadableCollectable(Named, Readable, Collectable, WritesStreamAssets):
    """Produces no events, but only StreamResources for 2 data keys and can be read or collected"""

    describe = describe_stream_datum
    describe_collect = describe_stream_datum
    read = read_empty
    collect_asset_docs = collect_asset_docs_stream_datum
    get_index = get_index


def test_datum_readable_counts(RE):
    """Test that count-ing a datum-producing device results in expected documents."""
    det = DatumReadable(name="det")
    docs = DocHolder()
    RE(bp.count([det]), docs.append)
    docs.assert_emitted(start=1, descriptor=1, resource=1, datum=1, event=1, stop=1)
    assert docs["descriptor"][0]["name"] == "primary"
    assert list(docs["descriptor"][0]["data_keys"]) == ["det-datum"]
    assert docs["event"][0]["descriptor"] == docs["descriptor"][0]["uid"]
    assert docs["event"][0]["data"] == {"det-datum": "RESOURCEUID/1"}
    assert docs["event"][0]["timestamps"] == {"det-datum": 456}
    assert docs["event"][0]["filled"] == {"det-datum": False}


def test_stream_datum_readable_counts(RE):
    """Test that count-ing a StreamDatum-producing device results in expected documents."""
    det = StreamDatumReadableCollectable(name="det")
    docs = DocHolder()
    RE(bp.count([det], 2), docs.append)
    docs.assert_emitted(start=1, descriptor=1, stream_resource=2, stream_datum=4, event=2, stop=1)
    assert docs["descriptor"][0]["name"] == "primary"
    assert list(docs["descriptor"][0]["data_keys"]) == ["det-sd1", "det-sd2"]
    assert all(sd["descriptor"] == docs["descriptor"][0]["uid"] for sd in docs["stream_datum"])
    assert all(e["data"] == {} for e in docs["event"])
    assert all(e["filled"] == {} for e in docs["event"])
    assert [sd["indices"] for sd in docs["stream_datum"]] == [
        {"start": 0, "stop": 1},
        {"start": 0, "stop": 1},
        {"start": 1, "stop": 2},
        {"start": 1, "stop": 2},
    ]


def test_combinations_counts(RE):
    """Test that mixing a StreamDatum- and Datum-producing device in one count works."""
    det1 = PvAndDatumReadable(name="det1")
    det2 = PvAndStreamDatumReadable(name="det2")
    docs = DocHolder()
    RE(bp.count([det1, det2]), docs.append)
    docs.assert_emitted(
        start=1,
        descriptor=1,
        resource=1,
        datum=1,
        stream_resource=2,
        stream_datum=2,
        event=1,
        stop=1,
    )
    assert docs["descriptor"][0]["name"] == "primary"
    # TODO: only works with a set at the moment
    assert set(docs["descriptor"][0]["data_keys"]) == {
        "det1-pv",
        "det1-datum",
        "det2-pv",
        "det2-sd1",
        "det2-sd2",
    }
    assert docs["event"][0]["descriptor"] == docs["descriptor"][0]["uid"]
    assert docs["event"][0]["data"] == {
        "det1-datum": "RESOURCEUID/1",
        "det1-pv": 5.8,
        "det2-pv": 5.8,
    }
    assert docs["event"][0]["timestamps"] == {
        "det1-datum": 456,
        "det1-pv": 123,
        "det2-pv": 123,
    }
    assert docs["event"][0]["filled"] == {"det1-datum": False}


def test_collect_stream_true_raises(RE):
    with pytest.raises(
        RuntimeError,
        match=re.escape(
            "Collect now emits EventPages (stream=False), so emitting Events (stream=True) is no longer supported"
        ),
    ):
        RE(collect_plan(OldPvCollectable("det"), pre_declare=False, stream=True))


def test_predeclare_requires_stream_name(RE):
    with pytest.raises(
        AssertionError,
        match=re.escape("A stream name that is not None is required for pre-declare"),
    ):
        RE(collect_plan(OldPvCollectable(name="det"), pre_declare=True))


def test_new_style_with_steam_name_requires_pre_declare(RE):
    with pytest.raises(
        AssertionError,
        match=re.escape("If a message stream name is provided declare stream needs to be called first."),
    ):
        RE(collect_plan(StreamDatumReadableCollectable(name="det"), pre_declare=False, stream_name="main"))


def test_new_style_with_no_stream_name_and_no_pre_declare_does_not_try_and_make_a_stream(RE):
    with pytest.raises(
        AssertionError,
        match=re.escape("Single nested data keys should be pre-declared"),
    ):
        RE(collect_plan(StreamDatumReadableCollectable(name="det"), pre_declare=False))


def test_same_key_in_multiple_streams_fails(RE):
    with pytest.raises(
        RuntimeError,
        match=re.escape("Can't use identical data keys in multiple streams"),
    ):
        RE(collect_plan(MultiKeyOldCollectable(name="det"), pre_declare=False))


def test_old_pv_collectable(RE):
    det = OldPvCollectable(name="det")
    docs = DocHolder()
    RE(collect_plan(det, pre_declare=False), docs.append)
    docs.assert_emitted(start=1, descriptor=2, event_page=2, stop=1)
    assert [d["name"] for d in docs["descriptor"]] == ["stream1", "stream2"]
    assert list(docs["descriptor"][0]["data_keys"]) == ["stream1-pv1", "stream1-pv2"]
    assert list(docs["descriptor"][1]["data_keys"]) == ["stream2-pv1", "stream2-pv2"]
    assert docs["event_page"][0]["data"] == {
        "stream1-pv1": [0, 1],
        "stream1-pv2": [0, 1],
    }
    assert docs["event_page"][0]["timestamps"] == {
        "stream1-pv1": [100, 101],
        "stream1-pv2": [100, 101],
    }
    assert docs["event_page"][1]["data"] == {
        "stream2-pv1": [0, 1],
        "stream2-pv2": [0, 1],
    }


def test_old_datum_collectable(RE):
    det = OldDatumCollectable(name="det")
    docs = DocHolder()
    RE(collect_plan(det, pre_declare=False), docs.append)
    docs.assert_emitted(start=1, descriptor=1, resource=1, datum=1, event_page=1, stop=1)
    assert docs["descriptor"][0]["name"] == "stream3"
    assert list(docs["descriptor"][0]["data_keys"]) == ["det-datum"]
    assert docs["event_page"][0]["descriptor"] == docs["descriptor"][0]["uid"]
    assert docs["event_page"][0]["data"] == {"det-datum": ["RESOURCEUID/1"]}
    assert docs["event_page"][0]["timestamps"] == {"det-datum": [456]}
    assert docs["event_page"][0]["filled"] == {"det-datum": [False]}


def test_old_datum_and_pv_collectable(RE):
    det = OldPvAndDatumCollectable(name="det")
    docs = DocHolder()
    RE(collect_plan(det, pre_declare=False), docs.append)
    docs.assert_emitted(start=1, descriptor=3, resource=1, datum=1, event_page=3, stop=1)
    assert [d["name"] for d in docs["descriptor"]] == ["stream1", "stream2", "stream3"]
    assert list(docs["descriptor"][0]["data_keys"]) == ["stream1-pv1", "stream1-pv2"]
    assert list(docs["descriptor"][2]["data_keys"]) == ["det-datum"]
    assert docs["event_page"][0]["data"] == {
        "stream1-pv1": [0, 1],
        "stream1-pv2": [0, 1],
    }
    assert docs["event_page"][1]["data"] == {
        "stream2-pv1": [0, 1],
        "stream2-pv2": [0, 1],
    }
    assert docs["event_page"][2]["data"] == {"det-datum": ["RESOURCEUID/1"]}


@pytest.mark.parametrize("cls", [PvCollectable, PvPageCollectable])
def test_pv_collectable(RE, cls):
    det = cls(name="det")
    docs = DocHolder()
    RE(collect_plan_no_stream_name_in_collect(det, pre_declare=True, stream_name="main"), docs.append)
    docs.assert_emitted(start=1, descriptor=1, event_page=1, stop=1)
    assert docs["descriptor"][0]["name"] == "main"
    assert list(docs["descriptor"][0]["data_keys"]) == ["det-pv1", "det-pv2"]
    assert docs["event_page"][0]["data"] == {
        "det-pv1": [0, 1],
        "det-pv2": [0, 1],
    }
    assert docs["event_page"][0]["timestamps"] == {
        "det-pv1": [100, 101],
        "det-pv2": [100, 101],
    }


def test_new_collect_needs_predeclare(RE):
    with pytest.raises(
        AssertionError,
        match="Single nested data keys should be pre-declared",
    ):
        RE(collect_plan(PvCollectable(name="det"), pre_declare=False))


def test_stream_datum_collectable(RE):
    det = StreamDatumReadableCollectable(name="det")
    docs = DocHolder()
    RE(collect_plan(det, pre_declare=True, stream_name="main"), docs.append)
    docs.assert_emitted(start=1, descriptor=1, stream_resource=2, stream_datum=2, stop=1)
    assert docs["descriptor"][0]["name"] == "main"
    assert list(docs["descriptor"][0]["data_keys"]) == ["det-sd1", "det-sd2"]
    assert docs["stream_resource"][0]["data_key"] == "det-sd1"
    assert docs["stream_resource"][1]["data_key"] == "det-sd2"
    assert docs["stream_datum"][0]["descriptor"] == docs["descriptor"][0]["uid"]
    assert docs["stream_datum"][1]["descriptor"] == docs["descriptor"][0]["uid"]


def test_stream_datum_collectable_no_stream_name(RE):
    det = StreamDatumReadableCollectable(name="det")
    docs = DocHolder()
    RE(collect_plan_no_stream_name_in_collect(det, pre_declare=True, stream_name="main"), docs.append)
    docs.assert_emitted(start=1, descriptor=1, stream_resource=2, stream_datum=2, stop=1)
    assert docs["descriptor"][0]["name"] == "main"
    assert list(docs["descriptor"][0]["data_keys"]) == ["det-sd1", "det-sd2"]
    assert docs["stream_resource"][0]["data_key"] == "det-sd1"
    assert docs["stream_resource"][1]["data_key"] == "det-sd2"
    assert docs["stream_datum"][0]["descriptor"] == docs["descriptor"][0]["uid"]
    assert docs["stream_datum"][1]["descriptor"] == docs["descriptor"][0]["uid"]


@pytest.mark.parametrize("cls1", [PvCollectable, PvPageCollectable])
@pytest.mark.parametrize("cls2", [PvCollectable, PvPageCollectable, StreamDatumReadableCollectable])
def test_many_collectables_fails(RE, cls1, cls2):
    """If there are multiple objects they must all WritesStreamAssests"""
    det1, det2 = cls1(name="det1"), cls2(name="det2")
    with pytest.raises(
        AssertionError,
        match=re.escape("does not implement all WritesStreamAssets methods"),
    ):
        RE(collect_plan(det1, det2, pre_declare=False))


def test_many_stream_datum_collectables(RE):
    """Test collecting from multiple StreamDatum-producing devices."""
    det1 = StreamDatumReadableCollectable(name="det1")
    det2 = StreamDatumReadableCollectable(name="det2")
    docs = DocHolder()
    RE(collect_plan(det1, det2, pre_declare=True, stream_name="main"), docs.append)
    docs.assert_emitted(start=1, descriptor=1, stream_resource=4, stream_datum=4, stop=1)
    data_keys = ["det1-sd1", "det1-sd2", "det2-sd1", "det2-sd2"]
    assert docs["descriptor"][0]["name"] == "main"
    assert set(docs["descriptor"][0]["data_keys"]) == set(data_keys)  # This only works in a set
    assert [d["data_key"] for d in docs["stream_resource"]] == data_keys
    assert all(d["descriptor"] == docs["descriptor"][0]["uid"] for d in docs["stream_datum"])


def tomo_plan(*objs):
    """An applied flyscanning example plan"""
    yield from bps.open_run()
    for name in ["flats", "darks", "projections"]:
        # projections is flyscan, others are step scan, so set collect accordingly
        yield from bps.declare_stream(*objs, name=name, collect=name == "projections")
    # take some flats, then darks
    yield from bps.collect(*objs, name="flats")
    yield from bps.collect(*objs, name="darks")
    # do the flyscan
    yield from bps.collect(*objs, name="projections")
    # take some flats at the end
    yield from bps.collect(*objs, name="flats")
    yield from bps.close_run()


def test_tomography_multi_stream_same_detectors(RE):
    """Test tomo_plan applied example"""
    det1 = StreamDatumReadableCollectable(name="det1")
    det2 = StreamDatumReadableCollectable(name="det2")
    docs = DocHolder()
    RE(tomo_plan(det1, det2), docs.append)
    # TODO: after https://github.com/bluesky/event-model/issues/296 this might be:
    #   stream_resource=12,  # one per stream per dataset per detector
    docs.assert_emitted(
        start=1,
        descriptor=3,  # one per stream
        stream_resource=4,  # one per dataset per detector
        stream_datum=16,  # 2 in flats, 1 in darks, one in projs, per dataset per detector
        stop=1,
    )
    data_keys = ["det1-sd1", "det1-sd2", "det2-sd1", "det2-sd2"]
    assert [d["name"] for d in docs["descriptor"]] == ["flats", "darks", "projections"]
    assert all(frozenset(d["data_keys"]) == frozenset(data_keys) for d in docs["descriptor"])
    assert [d["data_key"] for d in docs["stream_resource"]] == data_keys
    assert [d["descriptor"] for d in docs["stream_datum"]] == (
        [docs["descriptor"][0]["uid"]] * 4
        + [docs["descriptor"][1]["uid"]] * 4
        + [docs["descriptor"][2]["uid"]] * 4
        + [docs["descriptor"][0]["uid"]] * 4
    )


def change_conf_plan(*objs):
    """Re-emit a fresh EventDescriptor after the first Event."""
    yield from bps.open_run()
    for _ in range(2):
        yield from bps.declare_stream(*objs, name="main")
        yield from bps.collect(*objs)
    yield from bps.close_run()


def test_multiple_declare_in_same_stream(RE):
    """Test re-emitting an EventDescriptor while using StreamDatum-producing devices."""
    det1 = StreamDatumReadableCollectable(name="det1")
    det2 = StreamDatumReadableCollectable(name="det2")
    docs = DocHolder()
    RE(change_conf_plan(det1, det2), docs.append)
    docs.assert_emitted(start=1, descriptor=2, stream_resource=4, stream_datum=8, stop=1)
    data_keys = ["det1-sd1", "det1-sd2", "det2-sd1", "det2-sd2"]
    assert docs["descriptor"][0]["name"] == "main"
    assert frozenset(docs["descriptor"][0]["data_keys"]) == frozenset(data_keys)
    assert all(d["descriptor"] == docs["descriptor"][0]["uid"] for d in docs["stream_datum"][:4])
    assert docs["descriptor"][1]["name"] == "main"
    assert frozenset(docs["descriptor"][1]["data_keys"]) == frozenset(data_keys)
    assert all(d["descriptor"] == docs["descriptor"][1]["uid"] for d in docs["stream_datum"][4:])



import time
from collections import defaultdict
from enum import Enum
from io import StringIO
from itertools import permutations
from unittest.mock import MagicMock

import matplotlib.pyplot as plt
import numpy as np
import pytest
from event_model import DocumentNames, compose_run

import bluesky.plans as bp
import bluesky.preprocessors as bpp
from bluesky.callbacks import CallbackBase, CallbackCounter, LiveFit, LiveTable
from bluesky.callbacks.broker import BrokerCallbackBase
from bluesky.callbacks.core import make_callback_safe, make_class_safe
from bluesky.callbacks.mpl_plotting import LiveFitPlot, LiveGrid, LiveMesh, LivePlot, LiveRaster, LiveScatter
from bluesky.plan_stubs import pause
from bluesky.plans import count, grid_scan, inner_product_scan, scan
from bluesky.preprocessors import run_wrapper, subs_wrapper
from bluesky.run_engine import Msg, RunEngineInterrupted
from bluesky.tests.utils import DocCollector, MsgCollector, _print_redirect


# copied from examples.py to avoid import
def stepscan(det, motor):
    yield Msg("open_run")
    yield Msg("declare_stream", None, det, motor, name="primary")
    for i in range(-5, 5):
        yield Msg("create", name="primary")
        yield Msg("set", motor, i)
        yield Msg("trigger", det)
        yield Msg("read", motor)
        yield Msg("read", det)
        yield Msg("save")
    yield Msg("close_run")


def exception_raiser(name, doc):
    raise Exception("it's an exception that better not kill the scan!!")


def test_all(RE, hw):
    c = CallbackCounter()
    RE(stepscan(hw.det, hw.motor), {"all": c})
    assert c.value == 10 + 1 + 2  # events, descriptor, start and stop

    c = CallbackCounter()
    token = RE.subscribe(c)
    RE(stepscan(hw.det, hw.motor))
    RE.unsubscribe(token)
    assert c.value == 10 + 1 + 2


def test_raising_ignored_or_not(RE, hw):
    RE.ignore_callback_exceptions = True
    assert RE.ignore_callback_exceptions

    def cb(name, doc):
        raise Exception

    # by default (with ignore... = True) it warns
    with pytest.warns(UserWarning):
        RE(stepscan(hw.det, hw.motor), cb)

    RE.ignore_callback_exceptions = False
    with pytest.raises(Exception):  # noqa: B017
        RE(stepscan(hw.det, hw.motor), cb)


def test_subscribe_msg(RE, hw):
    assert RE.state == "idle"
    c = CallbackCounter()

    def counting_stepscan(det, motor):
        yield Msg("subscribe", None, c, "start")
        yield from stepscan(det, motor)

    RE(counting_stepscan(hw.det, hw.motor))  # should advance c
    assert c.value == 1
    RE(counting_stepscan(hw.det, hw.motor))  # should advance c
    assert c.value == 2
    RE(stepscan(hw.det, hw.motor))  # should not
    assert c.value == 2


def test_unknown_cb_raises(RE):
    def f(name, doc):
        pass

    with pytest.raises(KeyError):
        RE.subscribe(f, "not a thing")
    # back-compat alias for subscribe
    with pytest.raises(KeyError):
        RE.subscribe_lossless(f, "not a thing")
    with pytest.raises(KeyError):
        RE._subscribe_lossless(f, "not a thing")


def test_table_warns():
    table = LiveTable(["field"])
    table("start", {})
    with pytest.warns(UserWarning):
        table("descriptor", {"uid": "asdf", "name": "primary", "data_keys": {"field": {"dtype": "array"}}})


def test_table_external(RE, hw, db):
    RE.subscribe(db.insert)
    RE(count([hw.img]), LiveTable(["img"]))


def _compare_tables(fout, known_table):
    for ln, kn in zip(fout, known_table.split("\n")):
        # this is to strip the `\n` from the print output
        ln = ln.rstrip()
        if ln[0] == "+":
            # test the full line on the divider lines
            assert ln == kn
        else:
            # skip the 'time' column on data rows
            # this is easier than faking up times in the scan!
            assert ln[:16] == kn[:16]
            assert ln[26:] == kn[26:]


def test_table(RE, hw):
    with _print_redirect() as fout:
        hw.det.precision = 2
        hw.motor.precision = 2
        hw.motor.setpoint.put(0.0)  # Make dtype 'number' not 'integer'.
        hw.det.trigger()
        assert hw.det.describe()["det"]["precision"] == 2
        assert hw.motor.describe()["motor"]["precision"] == 2
        assert hw.det.describe()["det"]["dtype"] == "number"
        assert hw.motor.describe()["motor"]["dtype"] == "number"

        table = LiveTable(["det", "motor"], min_width=16, extra_pad=2, separator_lines=False)
        ad_scan = bp.adaptive_scan([hw.det], "det", hw.motor, -15.0, 5.0, 0.01, 1, 0.05, True)
        # use lossless sub here because rows can get dropped
        token = RE.subscribe(table)
        RE(ad_scan)
        RE.unsubscribe_lossless(token)

    fout.seek(0)
    _compare_tables(fout, KNOWN_TABLE)
    fout.close()


KNOWN_TABLE = """+------------+--------------+----------------+----------------+
|   seq_num  |        time  |           det  |         motor  |
+------------+--------------+----------------+----------------+
|         1  |  04:17:20.6  |          0.00  |        -15.00  |
|         2  |  04:17:20.6  |          0.00  |        -14.51  |
|         3  |  04:17:20.7  |          0.00  |        -13.91  |
|         4  |  04:17:20.8  |          0.00  |        -13.23  |
|         5  |  04:17:20.9  |          0.00  |        -12.49  |
|         6  |  04:17:20.9  |          0.00  |        -11.70  |
|         7  |  04:17:21.0  |          0.00  |        -10.86  |
|         8  |  04:17:21.1  |          0.00  |        -10.00  |
|         9  |  04:17:21.2  |          0.00  |         -9.10  |
|        10  |  04:17:21.2  |          0.00  |         -8.19  |
|        11  |  04:17:21.3  |          0.00  |         -7.25  |
|        12  |  04:17:21.4  |          0.00  |         -6.31  |
|        13  |  04:17:21.5  |          0.00  |         -5.35  |
|        14  |  04:17:21.5  |          0.00  |         -4.39  |
|        15  |  04:17:21.6  |          0.00  |         -3.41  |
|        16  |  04:17:21.7  |          0.05  |         -2.44  |
|        17  |  04:17:21.8  |          0.35  |         -1.45  |
|        18  |  04:17:21.8  |          0.08  |         -2.27  |
|        19  |  04:17:21.9  |          0.12  |         -2.07  |
|        20  |  04:17:22.0  |          0.18  |         -1.86  |
|        21  |  04:17:22.1  |          0.25  |         -1.66  |
|        22  |  04:17:22.1  |          0.22  |         -1.73  |
|        23  |  04:17:22.2  |          0.28  |         -1.59  |
|        24  |  04:17:22.3  |          0.34  |         -1.46  |
|        25  |  04:17:22.4  |          0.33  |         -1.49  |
|        26  |  04:17:22.4  |          0.38  |         -1.38  |
|        27  |  04:17:22.5  |          0.44  |         -1.28  |
|        28  |  04:17:22.6  |          0.50  |         -1.18  |
|        29  |  04:17:22.7  |          0.56  |         -1.08  |
|        30  |  04:17:22.7  |          0.62  |         -0.98  |
|        31  |  04:17:22.8  |          0.67  |         -0.89  |
|        32  |  04:17:22.9  |          0.73  |         -0.80  |
|        33  |  04:17:23.0  |          0.78  |         -0.71  |
|        34  |  04:17:23.1  |          0.82  |         -0.62  |
|        35  |  04:17:23.1  |          0.87  |         -0.53  |
|        36  |  04:17:23.2  |          0.91  |         -0.44  |
|        37  |  04:17:23.3  |          0.94  |         -0.34  |
|        38  |  04:17:23.4  |          0.97  |         -0.23  |
|        39  |  04:17:23.5  |          0.99  |         -0.11  |
|        40  |  04:17:23.5  |          1.00  |          0.04  |
|        41  |  04:17:23.6  |          0.94  |          0.36  |
|        42  |  04:17:23.7  |          0.96  |          0.30  |
|        43  |  04:17:23.8  |          0.85  |          0.56  |
|        44  |  04:17:23.9  |          0.91  |          0.42  |
|        45  |  04:17:23.9  |          0.86  |          0.56  |
|        46  |  04:17:24.0  |          0.79  |          0.69  |
|        47  |  04:17:24.1  |          0.81  |          0.66  |
|        48  |  04:17:24.2  |          0.75  |          0.75  |
|        49  |  04:17:24.3  |          0.70  |          0.85  |
+------------+--------------+----------------+----------------+
|   seq_num  |        time  |           det  |         motor  |
+------------+--------------+----------------+----------------+
|        50  |  04:17:24.3  |          0.64  |          0.94  |
|        51  |  04:17:24.4  |          0.58  |          1.04  |
|        52  |  04:17:24.5  |          0.53  |          1.13  |
|        53  |  04:17:24.6  |          0.48  |          1.22  |
|        54  |  04:17:24.7  |          0.43  |          1.30  |
|        55  |  04:17:24.7  |          0.38  |          1.39  |
|        56  |  04:17:24.8  |          0.33  |          1.48  |
|        57  |  04:17:24.9  |          0.29  |          1.57  |
|        58  |  04:17:25.0  |          0.25  |          1.66  |
|        59  |  04:17:25.0  |          0.21  |          1.76  |
|        60  |  04:17:25.1  |          0.18  |          1.87  |
|        61  |  04:17:25.2  |          0.14  |          1.98  |
|        62  |  04:17:25.3  |          0.11  |          2.10  |
|        63  |  04:17:25.4  |          0.08  |          2.24  |
|        64  |  04:17:25.4  |          0.06  |          2.39  |
|        65  |  04:17:25.5  |          0.04  |          2.58  |
|        66  |  04:17:25.6  |          0.02  |          2.82  |
|        67  |  04:17:25.7  |          0.01  |          3.16  |
|        68  |  04:17:25.8  |          0.00  |          3.62  |
|        69  |  04:17:25.8  |          0.00  |          4.20  |
|        70  |  04:17:25.9  |          0.00  |          4.85  |
+------------+--------------+----------------+----------------+"""


def test_evil_table_names(RE):
    from ophyd import Signal

    sigs = [
        Signal(value=0, name="a:b"),
        Signal(value=0, name="a,b"),
        Signal(value=0, name="a'b"),
        Signal(value=0, name="🐍"),
    ]
    table = LiveTable([s.name for s in sigs], min_width=5, extra_pad=2, separator_lines=False)
    with _print_redirect() as fout:
        print()  # get a blank line in camptured output
        RE(bpp.subs_wrapper(bp.count(sigs, num=2), table))
    reference = """
+------------+--------------+--------+--------+--------+--------+
|   seq_num  |        time  |   a:b  |   a,b  |   a'b  |     🐍  |
+------------+--------------+--------+--------+--------+--------+
|         1  |  12:47:09.7  |     0  |     0  |     0  |     0  |
|         2  |  12:47:09.7  |     0  |     0  |     0  |     0  |
+------------+--------------+--------+--------+--------+--------+"""
    _compare_tables(fout, reference)
    fout.close()


class TableTestEnum(str, Enum):
    """Enum for testing LiveTable with enum values."""

    OK = "OK"
    NOT_OK = "NOT_OK"


_BOOL_ENUM_DOCS = [
    (
        "descriptor",
        {
            "configuration": {},
            "data_keys": {
                "enum_test_sig": {
                    "dtype": "string",
                    "shape": [],
                    "dtype_numpy": "|S40",
                    "source": "ca://DEV:RBD9103:InRange_RBV",
                    "choices": ["OK", "Under", "Over"],
                    "object_name": "rbd9103",
                },
                "bool_test_sig": {
                    "dtype": "boolean",
                    "shape": [],
                    "dtype_numpy": "|b1",
                    "source": "ca://DEV:RBD9103:Stable_RBV",
                    "object_name": "rbd9103",
                },
            },
            "name": "primary",
            "object_keys": {"rbd9103": ["enum_test_sig", "bool_test_sig"]},
            "run_start": "d8e2afc6-2c79-4786-8a2e-b1245b2be578",
            "time": 1746210217.2738924,
            "uid": "2d29452a-a2e3-4697-8db2-423b8cae5065",
            "hints": {"rbd9103": {"fields": ["enum_test_sig", "bool_test_sig"]}},
        },
    ),
    (
        "event",
        {
            "uid": "69b921f3-6037-477b-8d76-731c330c654a",
            "time": 1746210217.2745216,
            "data": {"enum_test_sig": TableTestEnum.OK, "bool_test_sig": False},
            "timestamps": {"enum_test_sig": 631152000.0, "bool_test_sig": 631152000.0},
            "seq_num": 1,
            "filled": {},
            "descriptor": "2d29452a-a2e3-4697-8db2-423b8cae5065",
        },
    ),
    (
        "event",
        {
            "uid": "69b921f3-6037-477b-8d76-731c330c654a",
            "time": 1746210217.2745216,
            "data": {"enum_test_sig": TableTestEnum.NOT_OK, "bool_test_sig": True},
            "timestamps": {"enum_test_sig": 631152000.0, "bool_test_sig": 631152000.0},
            "seq_num": 2,
            "filled": {},
            "descriptor": "2d29452a-a2e3-4697-8db2-423b8cae5065",
        },
    ),
    (
        "stop",
        {
            "uid": "0c862732-f3a2-4ee6-8ef8-7aed0cf4a408",
            "time": 1746210217.2791321,
            "run_start": "d8e2afc6-2c79-4786-8a2e-b1245b2be578",
            "exit_status": "success",
            "reason": "",
            "num_events": {"primary": 1},
        },
    ),
]


def test_table_bool_enum_doc_inputs(RE):
    table = LiveTable(list(_BOOL_ENUM_DOCS[0][1]["data_keys"]))
    with _print_redirect() as fout:
        print()  # get a blank line in camptured output
        for name, doc in _BOOL_ENUM_DOCS:
            table(name, doc)
    reference = """
+-----------+------------+---------------+---------------+
|   seq_num |       time | enum_test_sig | bool_test_sig |
+-----------+------------+---------------+---------------+
|         1 | 14:23:37.2 |            OK |         False |
|         2 | 14:23:37.2 |        NOT_OK |          True |
+-----------+------------+---------------+---------------+"""
    _compare_tables(fout, reference)
    fout.close()


def test_live_fit(RE, hw):
    try:
        import lmfit
    except ImportError:
        raise pytest.skip("requires lmfit")  # noqa: B904

    def gaussian(x, A, sigma, x0):
        return A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))

    model = lmfit.Model(gaussian)
    init_guess = {"A": 2, "sigma": lmfit.Parameter("sigma", 3, min=0), "x0": -0.2}
    cb = LiveFit(model, "det", {"x": "motor"}, init_guess, update_every=50)
    RE(scan([hw.det], hw.motor, -1, 1, 50), cb)
    # results are in cb.result.values

    expected = {"A": 1, "sigma": 1, "x0": 0}
    for k, v in expected.items():
        assert np.allclose(cb.result.values[k], v, atol=1e-6)


def test_live_fit_multidim(RE, hw):
    try:
        import lmfit
    except ImportError:
        raise pytest.skip("requires lmfit")  # noqa: B904

    hw.motor1.delay = 0
    hw.motor2.delay = 0
    hw.det4.exposure_time = 0

    def gaussian(x, y, A, sigma, x0, y0):
        return A * np.exp(-((x - x0) ** 2 + (y - y0) ** 2) / (2 * sigma**2))

    model = lmfit.Model(gaussian, ["x", "y"])
    init_guess = {"A": 2, "sigma": lmfit.Parameter("sigma", 3, min=0), "x0": -0.2, "y0": 0.3}
    cb = LiveFit(model, "det4", {"x": "motor1", "y": "motor2"}, init_guess, update_every=50)
    RE(grid_scan([hw.det4], hw.motor1, -1, 1, 10, hw.motor2, -1, 1, 10, False), cb)

    expected = {"A": 1, "sigma": 1, "x0": 0, "y0": 0}
    for k, v in expected.items():
        assert np.allclose(cb.result.values[k], v, atol=1e-6)


def test_live_plot_from_callbacks():
    import bluesky.callbacks.core

    # We don't want the shims in callbacks.core, see #1133 for discussion
    assert not hasattr(bluesky.callbacks.core, "LivePlot")
    # We still want the shims in callbacks.__init__
    from bluesky.callbacks import LivePlot as LivePlotFromCallbacks

    assert LivePlotFromCallbacks is LivePlot

    # Make sure we can subclass it
    class FromCallbacks(LivePlotFromCallbacks): ...

    FromCallbacks("det", "motor")


def test_live_fit_plot(RE, hw):
    try:
        import lmfit
    except ImportError:
        raise pytest.skip("requires lmfit")  # noqa: B904

    def gaussian(x, A, sigma, x0):
        return A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))

    model = lmfit.Model(gaussian)
    init_guess = {"A": 2, "sigma": lmfit.Parameter("sigma", 3, min=0), "x0": -0.2}
    livefit = LiveFit(model, "det", {"x": "motor"}, init_guess, update_every=50)
    lfplot = LiveFitPlot(livefit, color="r")
    lplot = LivePlot("det", "motor", ax=plt.gca(), marker="o", ls="none")
    RE(scan([hw.det], hw.motor, -1, 1, 50), [lplot, lfplot])
    expected = {"A": 1, "sigma": 1, "x0": 0}
    for k, v in expected.items():
        assert np.allclose(livefit.result.values[k], v, atol=1e-6)


@pytest.mark.parametrize("int_meth, stop_num, msg_num", [("stop", 1, 5), ("abort", 1, 5), ("halt", 1, 3)])
def test_interrupted_with_callbacks(RE, int_meth, stop_num, msg_num):
    docs = defaultdict(list)

    def collector_cb(name, doc):
        nonlocal docs
        docs[name].append(doc)

    RE.msg_hook = MsgCollector()

    with pytest.raises(RunEngineInterrupted):
        RE(subs_wrapper(run_wrapper(pause()), {"all": collector_cb}))
    getattr(RE, int_meth)()

    assert len(docs["start"]) == 1
    assert len(docs["event"]) == 0
    assert len(docs["descriptor"]) == 0
    assert len(docs["stop"]) == stop_num
    assert len(RE.msg_hook.msgs) == msg_num


def test_live_grid(RE, hw):
    hw.motor1.delay = 0
    hw.motor2.delay = 0
    RE(grid_scan([hw.det4], hw.motor1, -3, 3, 6, hw.motor2, -5, 5, 10, False), LiveGrid((6, 10), "det4"))

    # Test the deprecated name.
    with pytest.warns(UserWarning):
        RE(grid_scan([hw.det4], hw.motor1, -3, 3, 6, hw.motor2, -5, 5, 10, False), LiveRaster((6, 10), "det4"))


def test_live_grid_title(RE, hw):
    hw.motor1.delay = 0
    hw.motor2.delay = 0
    RE(
        grid_scan([hw.det4], hw.motor1, -3, 3, 6, hw.motor2, -5, 5, 10, False),
        LiveGrid((6, 10), "det4", title="Det 4 Grid Plot"),
    )


def test_live_scatter(RE, hw):
    RE(
        grid_scan([hw.det5], hw.jittery_motor1, -3, 3, 6, hw.jittery_motor2, -5, 5, 10, False),
        LiveScatter("jittery_motor1", "jittery_motor2", "det5", xlim=(-3, 3), ylim=(-5, 5)),
    )

    # Test the deprecated name.
    with pytest.warns(UserWarning):
        RE(
            grid_scan([hw.det5], hw.jittery_motor1, -3, 3, 6, hw.jittery_motor2, -5, 5, 10, False),
            LiveMesh("jittery_motor1", "jittery_motor2", "det5", xlim=(-3, 3), ylim=(-5, 5)),
        )


def test_live_scatter_auto_range(RE, hw):
    cb = LiveScatter("motor1", "motor2", "det5")
    RE(grid_scan([hw.det5], hw.motor1, -3, 3, 6, hw.motor2, -5, 5, 10, False), cb)
    assert cb.minx is not None
    assert cb.maxx is not None
    assert cb.miny is not None
    assert cb.maxy is not None
    assert cb.ax.get_xlim()[0] == -3
    assert cb.ax.get_xlim()[1] == 3
    assert cb.ax.get_ylim()[0] == -5
    assert cb.ax.get_ylim()[1] == 5


@pytest.mark.xfail(
    raises=NotImplementedError, reason="This tests an API databroker has removed, and needs updating."
)
def test_broker_base(RE, hw, db):
    class BrokerChecker(BrokerCallbackBase):
        def __init__(self, field, *, db=None):
            super().__init__(field, db=db)

        def event(self, doc):
            super().event(doc)
            assert isinstance(doc["data"][self.fields[0]], np.ndarray)

    RE.subscribe(db.insert)
    bc = BrokerChecker(("img",), db=db)
    RE.subscribe(bc)
    RE(count([hw.img]))


def test_broker_base_no_unpack(RE, hw, db):
    class BrokerChecker(BrokerCallbackBase):
        def __init__(self, field, *, db=None):
            super().__init__(field, db=db)

        def event(self, doc):
            super().event(doc)
            assert isinstance(doc["data"][self.fields[0]], np.ndarray)

    bc = BrokerChecker(("img",), db=db)
    RE.subscribe(bc)
    RE(count([hw.direct_img]))


def test_plotting_hints(RE, hw, db):
    """This tests the run and checks that the correct hints are created.
    Hints are mainly created to help the BestEffortCallback in plotting the
    data.
    Use a callback to do the checking.
    """
    dc = DocCollector()
    RE.subscribe(dc.insert)

    # check that the inner product hints are passed correctly
    hint = {"dimensions": [([hw.motor1.name, hw.motor2.name, hw.motor3.name], "primary")]}
    RE(inner_product_scan([hw.det], 20, hw.motor1, -1, 1, hw.motor2, -1, 1, hw.motor3, -2, 0))
    assert dc.start[-1]["hints"] == hint

    # check that the outer product (grid_scan) hints are passed correctly
    hint = {"dimensions": [(["motor1"], "primary"), (["motor2"], "primary"), (["motor3"], "primary")]}
    # grid_scan passes "rectilinear" gridding as well
    # make sure this is also passed
    output_hint = hint.copy()
    output_hint["gridding"] = "rectilinear"
    RE(grid_scan([hw.det], hw.motor1, -1, 1, 2, hw.motor2, -1, 1, 2, True, hw.motor3, -2, 0, 2, True))

    assert dc.start[-1]["hints"] == output_hint

    # check that if gridding is supplied, it's not overwritten by grid_scan
    # check that the outer product (grid_scan) hints are passed correctly
    hint = {
        "dimensions": [(["motor1"], "primary"), (["motor2"], "primary"), (["motor3"], "primary")],
        "gridding": "rectilinear",
    }
    RE(grid_scan([hw.det], hw.motor1, -1, 1, 2, hw.motor2, -1, 1, 2, True, hw.motor3, -2, 0, 2, True))
    assert dc.start[-1]["hints"] == hint


def test_broken_table():
    start_doc, descriptor_factory, *_ = compose_run()
    desc, compose_event, _ = descriptor_factory(
        name="primary",
        data_keys={
            "x": {"dtype": "integer", "source": "", "shape": []},
            "y": {"dtype": "number", "source": "", "shape": []},
        },
    )

    ev1 = compose_event(data={"x": 1, "y": 2.0}, timestamps={k: time.time() for k in ("x", "y")})
    ev2 = compose_event(data={"x": 1, "y": 2}, timestamps={k: time.time() for k in ("x", "y")})
    ev3 = compose_event(data={"x": 1.0, "y": 2.0}, timestamps={k: time.time() for k in ("x", "y")})
    ev4 = compose_event(
        data={"x": 1.0, "y": "aardvark"},
        timestamps={k: time.time() for k in ("x", "y")},
    )

    sio = StringIO()
    LT = LiveTable(["x", "y"], out=lambda s: sio.write(s + "\n"))
    LT("start", start_doc)
    LT("descriptor", desc)
    LT("event", ev1)
    LT("event", ev2)
    LT("event", ev3)
    LT("event", ev4)

    sio.seek(0)
    lines = sio.readlines()

    # The instance of LiveTable will include two empty separator lines by default
    assert len(lines) == 9
    for ln in lines[-2:]:
        assert ln.strip() == "failed to format row"


def test_callback_safe():
    @make_callback_safe
    def test_function(to_fail):
        if to_fail:
            raise RuntimeError
        return to_fail

    assert test_function(True) is None
    assert test_function(False) is False


def test_callback_safe_logger():
    from types import SimpleNamespace
    from unittest.mock import MagicMock

    logger = SimpleNamespace(exception=MagicMock())

    @make_callback_safe(logger=logger)
    def test_function(to_fail):
        if to_fail:
            raise RuntimeError
        return to_fail

    assert test_function(True) is None
    assert logger.exception.call_count == 1
    assert test_function(False) is False
    assert logger.exception.call_count == 1


@pytest.fixture
def EvilBaseClass(request):
    class MyError(RuntimeError): ...

    class EvilCallback(CallbackBase):
        my_excepttion_type = MyError

        def event(self, doc):
            raise MyError

        def bulk_events(self, doc):
            raise MyError

        def resource(self, doc):
            raise MyError

        def datum(self, doc):
            raise MyError

        def bulk_datum(self, doc):
            raise MyError

        def descriptor(self, doc):
            raise MyError

        def start(self, doc):
            raise MyError

        def stop(self, doc):
            raise MyError

        def event_page(self, doc):
            raise MyError

        def datum_page(self, doc):
            raise MyError

        def stream_resource(self, doc):
            raise MyError

        def stream_datum(self, doc):
            raise MyError

    return EvilCallback


def test_callbackclass(EvilBaseClass):
    ecb = EvilBaseClass()
    for n in DocumentNames:
        with pytest.raises(EvilBaseClass.my_excepttion_type):
            ecb(n.name, {})


def test_callbackclass_safe(EvilBaseClass):
    @make_class_safe
    class SafeEvilBaseClass(EvilBaseClass): ...

    scb = SafeEvilBaseClass()
    for n in DocumentNames:
        scb(n.name, {})


def test_callbackclass_safe_logger(EvilBaseClass):
    logger = MagicMock()

    @make_class_safe(logger=logger)
    class SafeEvilBaseClass2(EvilBaseClass): ...

    scb = SafeEvilBaseClass2()
    for n in DocumentNames:
        scb(n.name, {})

    assert logger.exception.call_count == len(DocumentNames)


@pytest.mark.parametrize("strict", ["1", None])
@pytest.mark.parametrize(
    "documents",
    (
        list(set(tuple(sorted(x, key=lambda x: x.name)) for x in permutations(DocumentNames, 1)))  # noqa: C401
        + list(set(tuple(sorted(x, key=lambda x: x.name)) for x in permutations(DocumentNames, 2)))  # noqa: C401
        + list(set(tuple(sorted(x, key=lambda x: x.name)) for x in permutations(DocumentNames, 3)))  # noqa: C401
        + [list(DocumentNames)]
    ),
)
def test_callbackclass_safe_filtered(EvilBaseClass, documents, monkeypatch, strict):
    if strict is not None:
        monkeypatch.setenv("BLUESKY_DEBUG_CALLBACKS", strict)
    else:
        monkeypatch.delenv("BLUESKY_DEBUG_CALLBACKS", raising=False)
    logger = MagicMock()

    @make_class_safe(logger=logger, to_wrap=tuple(x.name for x in documents))
    class SafeEvilBaseClass2(EvilBaseClass): ...

    scb = SafeEvilBaseClass2()
    for n in documents:
        if strict:
            with pytest.raises(EvilBaseClass.my_excepttion_type):
                scb(n.name, {})
        else:
            scb(n.name, {})

    for n in set(DocumentNames) - set(documents):
        with pytest.raises(EvilBaseClass.my_excepttion_type):
            scb(n.name, {})

    assert logger.exception.call_count == len(documents)


def test_in_plan_qt_callback(RE, hw):
    from bluesky.callbacks.mpl_plotting import _get_teleporter

    _get_teleporter()

    def my_plan():
        motor = hw.motor
        det = hw.det

        motor.delay = 1

        plan = bp.scan([det], motor, -5, 5, 25)
        plan = subs_wrapper(bp.scan([det], motor, -5, 5, 25), LivePlot(det.name, motor.name))
        return (yield from plan)

    RE(my_plan())
